/*
 * Copyright 1998  The XFree86 Project
 *
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the "Software"), 
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL 
 * JOERG POMMNITZ BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF 
 * OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
 * SOFTWARE.
 * 
 */

/*
 * XLFD field      | Source
 * ----------------+-----------------------------------------------
 * FOUNDRY         | "OS2 " table, achVendID field
 * FAMILY_NAME     | "name" table, Font Family name
 *                 | "name" table, Postscript name
 * WEIGHT_NAME     | "OS2 " table, fsSelection field bit 5
 *                 | "OS2 " table, PANOSE WEIGHT field
 *                 | "OS2 " table, usWeightClass field
 * SLANT           | "OS2 " table, fsSelection field bit 0
 * SETWIDTH_NAME   | "OS2 " table, usWidthClass field
 *                 | "OS2 " table, PANOSE proportion/aspect field
 * ADD_STYLE_NAME  | empty
 * PIXEL_SIZE      | 0
 * POINT_SIZE      | 0
 * RESOLUTION_X    | 0
 * RESOLUTION_Y    | 0
 * SPACING         | constant for CJK encodings
 *                 | "post" table, isFixedPitch flag
 * AVERAGE_WIDTH   | 0
 * CHARSET_REGISTRY| internal table
 * CHARSET_ENCODING| internal table, test for characters
 */

/*
 * Contributions:
 *   Pablo Saratxaga      : - support for more encodings, japanese (jisx201,jisx208),
 *                            chinese (gb2312, big5), korean (kcs5601) and iso-8859-10.
 *			      vietnamese (tcvn-5712, viscii1.1-1), lao, thai,
 *			      armenian (armscii-8), georgian
 *			      lithuanina (8859-13), celtic (8859-14)
 *                          - enhanced font name lookup 
 */

#include <unistd.h>
#include <sys/types.h>
#include <dirent.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>

#include <freetype/freetype.h>

#include "panose.h"
#include "ttos2val.h"

#define TTMKFS_USAGE  1
#define TTMKFS_FTINIT 2
#define TTMKFS_DIR    3
#define TTMKFS_MEM    4
#define TTMKFS_NO_UC  5

#define DEFAULTDIR "."
#define DEFAULTFOUNDRY "misc"

static char *path = NULL;            /* directory with the TTF files, by default the DEFAULTDIR */
static char *default_foundry = "misc"; /* the foundry name we can fall back to if we can't figure one out */
static FILE *outstream = NULL;       /* the stream we write our results to */
static int  always_unicode = 0;
static int  max_missing = 3;
static int  use_panose = 0;
static int  create_debug_entries = 0;
static int  count_encodings = 0; /* if larger than COUNT_ENCODINGS_UNICODE an -iso10646-1 entry will be produced; not all encodings are counted */
#define COUNT_ENCODINGS_UNICODE  10

unsigned long int entries;
struct xlfdmapping {
    char mapping[512];
    struct xlfdmapping *next;
};

enum completeness {
    IGNORE,
    STRICT,
    LOOSE
};

static int   parse_argv        (int argc, char *argv[]);
static void  usage             (const char *name);
static int   is_ttf            (const struct dirent *);
static const char *get_foundry (const TT_Char orig_code[4]);
static const char *get_name    (TT_Face face);
static void  add_mapping (struct xlfdmapping **listhead,
			  char *file_name,
			  const char *foundry_name,
			  const char *font_name,
			  const char *weight_name,
			  const char *widht_name,
			  TT_Face_Properties *properties,
			  char *encodingscheme
			  );

static int   has_8bit_encoding (TT_CharMap *cmap, const int *table);
static const char *get_weight (const char *name, TT_Face_Properties *properties);
static const char *get_width_name (const char *name, TT_Face_Properties *properties);


static enum completeness completeness = STRICT;

#define ALTCHR  -1
#define EMPTYLINE  \
        ALTCHR, ALTCHR, ALTCHR, ALTCHR, ALTCHR, ALTCHR, ALTCHR, ALTCHR \
        ALTCHR, ALTCHR, ALTCHR, ALTCHR, ALTCHR, ALTCHR, ALTCHR, ALTCHR

static int iso8859_1[224] = {
    0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027, 
    0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f, 
    0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037, 
    0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f, 
    0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047, 
    0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f, 
    0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057, 
    0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x005e, 0x005f, 
    0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067, 
    0x0068, 0x0069, 0x006a, 0x006b, 0x006c, 0x006d, 0x006e, 0x006f, 
    0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077, 
    0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d, 0x007e,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    0x00a0, 0x00a1, 0x00a2, 0x00a3, 0x00a4, 0x00a5, 0x00a6, 0x00a7, 
    0x00a8, 0x00a9, 0x00aa, 0x00ab, 0x00ac, 0x00ad, 0x00ae, 0x00af, 
    0x00b0, 0x00b1, 0x00b2, 0x00b3, 0x00b4, 0x00b5, 0x00b6, 0x00b7, 
    0x00b8, 0x00b9, 0x00ba, 0x00bb, 0x00bc, 0x00bd, 0x00be, 0x00bf, 
    0x00c0, 0x00c1, 0x00c2, 0x00c3, 0x00c4, 0x00c5, 0x00c6, 0x00c7, 
    0x00c8, 0x00c9, 0x00ca, 0x00cb, 0x00cc, 0x00cd, 0x00ce, 0x00cf, 
    0x00d0, 0x00d1, 0x00d2, 0x00d3, 0x00d4, 0x00d5, 0x00d6, 0x00d7, 
    0x00d8, 0x00d9, 0x00da, 0x00db, 0x00dc, 0x00dd, 0x00de, 0x00df, 
    0x00e0, 0x00e1, 0x00e2, 0x00e3, 0x00e4, 0x00e5, 0x00e6, 0x00e7, 
    0x00e8, 0x00e9, 0x00ea, 0x00eb, 0x00ec, 0x00ed, 0x00ee, 0x00ef, 
    0x00f0, 0x00f1, 0x00f2, 0x00f3, 0x00f4, 0x00f5, 0x00f6, 0x00f7, 
    0x00f8, 0x00f9, 0x00fa, 0x00fb, 0x00fc, 0x00fd, 0x00fe, 0x00ff
};

static int iso8859_2[224] = {
    0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027, 
    0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f, 
    0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037, 
    0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f, 
    0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047, 
    0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f, 
    0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057, 
    0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x005e, 0x005f, 
    0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067, 
    0x0068, 0x0069, 0x006a, 0x006b, 0x006c, 0x006d, 0x006e, 0x006f, 
    0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077, 
    0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d, 0x007e,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    0x00a0, 0x0104, 0x02d8, 0x0141, 0x00a4, 0x013d, 0x015a, 0x00a7, 
    0x00a8, 0x0160, 0x015e, 0x0164, 0x0179, 0x00ad, 0x017d, 0x017b, 
    0x00b0, 0x0105, 0x02db, 0x0142, 0x00b4, 0x013e, 0x015b, 0x02c7, 
    0x00b8, 0x0161, 0x015f, 0x0165, 0x017a, 0x02dd, 0x017e, 0x017c, 
    0x0154, 0x00c1, 0x00c2, 0x0102, 0x00c4, 0x0139, 0x0106, 0x00c7, 
    0x010c, 0x00c9, 0x0118, 0x00cb, 0x011a, 0x00cd, 0x00ce, 0x010e, 
    0x0110, 0x0143, 0x0147, 0x00d3, 0x00d4, 0x0150, 0x00d6, 0x00d7, 
    0x0158, 0x016e, 0x00da, 0x0170, 0x00dc, 0x00dd, 0x0162, 0x00df, 
    0x0155, 0x00e1, 0x00e2, 0x0103, 0x00e4, 0x013a, 0x0107, 0x00e7, 
    0x010d, 0x00e9, 0x0119, 0x00eb, 0x011b, 0x00ed, 0x00ee, 0x010f, 
    0x0111, 0x0144, 0x0148, 0x00f3, 0x00f4, 0x0151, 0x00f6, 0x00f7, 
    0x0159, 0x016f, 0x00fa, 0x0171, 0x00fc, 0x00fd, 0x0163, 0x02d9 
};

static int iso8859_3[224] = {
    0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027, 
    0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f, 
    0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037, 
    0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f, 
    0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047, 
    0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f, 
    0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057, 
    0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x005e, 0x005f, 
    0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067, 
    0x0068, 0x0069, 0x006a, 0x006b, 0x006c, 0x006d, 0x006e, 0x006f, 
    0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077, 
    0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d, 0x007e,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    0x00a0, 0x0126, 0x02d8, 0x00a3, 0x00a4,     -1, 0x0124, 0x00a7, 
    0x00a8, 0x0130, 0x015e, 0x011e, 0x0134, 0x00ad,     -1, 0x017b, 
    0x00b0, 0x0127, 0x00b2, 0x00b3, 0x00b4, 0x00b5, 0x0125, 0x00b7, 
    0x00b8, 0x0131, 0x015f, 0x011f, 0x0135, 0x00bd,     -1, 0x017c, 
    0x00c0, 0x00c1, 0x00c2,     -1, 0x00c4, 0x010a, 0x0108, 0x00c7, 
    0x00c8, 0x00c9, 0x00ca, 0x00cb, 0x00cc, 0x00cd, 0x00ce, 0x00cf, 
    /**/-1, 0x00d1, 0x00d2, 0x00d3, 0x00d4, 0x0120, 0x00d6, 0x00d7, 
    0x011c, 0x00d9, 0x00da, 0x00db, 0x00dc, 0x016c, 0x015c, 0x00df, 
    0x00e0, 0x00e1, 0x00e2,     -1, 0x00e4, 0x010b, 0x0109, 0x00e7, 
    0x00e8, 0x00e9, 0x00ea, 0x00eb, 0x00ec, 0x00ed, 0x00ee, 0x00ef, 
    /**/-1, 0x00f1, 0x00f2, 0x00f3, 0x00f4, 0x0121, 0x00f6, 0x00f7, 
    0x011d, 0x00f9, 0x00fa, 0x00fb, 0x00fc, 0x016d, 0x015d, 0x02d9 
};

static int iso8859_4[224] = {
    0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027, 
    0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f, 
    0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037, 
    0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f, 
    0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047, 
    0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f, 
    0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057, 
    0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x005e, 0x005f, 
    0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067, 
    0x0068, 0x0069, 0x006a, 0x006b, 0x006c, 0x006d, 0x006e, 0x006f, 
    0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077, 
    0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d, 0x007e,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    0x00a0, 0x0104, 0x0138, 0x0156, 0x00a4, 0x0128, 0x013b, 0x00a7, 
    0x00a8, 0x0160, 0x0112, 0x0122, 0x0166, 0x00ad, 0x017d, 0x00af, 
    0x00b0, 0x0105, 0x02db, 0x0157, 0x00b4, 0x0129, 0x013c, 0x02c7, 
    0x00b8, 0x0161, 0x0113, 0x0123, 0x0167, 0x014a, 0x017e, 0x014b, 
    0x0100, 0x00c1, 0x00c2, 0x00c3, 0x00c4, 0x00c5, 0x00c6, 0x012e, 
    0x010c, 0x00c9, 0x0118, 0x00cb, 0x0116, 0x00cd, 0x00ce, 0x012a, 
    0x0110, 0x0145, 0x014c, 0x0136, 0x00d4, 0x00d5, 0x00d6, 0x00d7, 
    0x00d8, 0x0172, 0x00da, 0x00db, 0x00dc, 0x0168, 0x016a, 0x00df, 
    0x0101, 0x00e1, 0x00e2, 0x00e3, 0x00e4, 0x00e5, 0x00e6, 0x012f, 
    0x010d, 0x00e9, 0x0119, 0x00eb, 0x0117, 0x00ed, 0x00ee, 0x012b, 
    0x0111, 0x0146, 0x014d, 0x0137, 0x00f4, 0x00f5, 0x00f6, 0x00f7, 
    0x00f8, 0x0173, 0x00fa, 0x00fb, 0x00fc, 0x0169, 0x016b, 0x02d9 
};

static int iso8859_5[224] = {
    0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027, 
    0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f, 
    0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037, 
    0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f, 
    0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047, 
    0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f, 
    0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057, 
    0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x005e, 0x005f, 
    0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067, 
    0x0068, 0x0069, 0x006a, 0x006b, 0x006c, 0x006d, 0x006e, 0x006f, 
    0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077, 
    0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d, 0x007e,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    0x00a0, 0x0401, 0x0402, 0x0403, 0x0404, 0x0405, 0x0406, 0x0407, 
    0x0408, 0x0409, 0x040a, 0x040b, 0x040c, 0x00ad, 0x040e, 0x040f, 
    0x0410, 0x0411, 0x0412, 0x0413, 0x0414, 0x0415, 0x0416, 0x0417, 
    0x0418, 0x0419, 0x041a, 0x041b, 0x041c, 0x041d, 0x041e, 0x041f, 
    0x0420, 0x0421, 0x0422, 0x0423, 0x0424, 0x0425, 0x0426, 0x0427, 
    0x0428, 0x0429, 0x042a, 0x042b, 0x042c, 0x042d, 0x042e, 0x042f, 
    0x0430, 0x0431, 0x0432, 0x0433, 0x0434, 0x0435, 0x0436, 0x0437, 
    0x0438, 0x0439, 0x043a, 0x043b, 0x043c, 0x043d, 0x043e, 0x043f, 
    0x0440, 0x0441, 0x0442, 0x0443, 0x0444, 0x0445, 0x0446, 0x0447, 
    0x0448, 0x0449, 0x044a, 0x044b, 0x044c, 0x044d, 0x044e, 0x044f, 
    0x2116, 0x0451, 0x0452, 0x0453, 0x0454, 0x0455, 0x0456, 0x0457, 
    0x0458, 0x0459, 0x045a, 0x045b, 0x045c, 0x00a7, 0x045e, 0x045f 
};

static int iso8859_6[224] = {
    0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027, 
    0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f, 
    0x0660, 0x0661, 0x0662, 0x0663, 0x0664, 0x0665, 0x0666, 0x0667, 
    0x0668, 0x0669, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f, 
    0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047, 
    0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f, 
    0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057, 
    0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x005e, 0x005f, 
    0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067, 
    0x0068, 0x0069, 0x006a, 0x006b, 0x006c, 0x006d, 0x006e, 0x006f, 
    0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077, 
    0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d, 0x007e,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    0x00a0,     -1,     -1,     -1, 0x00a4,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1, 0x060c, 0x00ad,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1, 0x061b,     -1,     -1,     -1, 0x061f, 
    /**/-1, 0x0621, 0x0622, 0x0623, 0x0624, 0x0625, 0x0626, 0x0627, 
    0x0628, 0x0629, 0x062a, 0x062b, 0x062c, 0x062d, 0x062e, 0x062f, 
    0x0630, 0x0631, 0x0632, 0x0633, 0x0634, 0x0635, 0x0636, 0x0637, 
    0x0638, 0x0639, 0x063a,     -1,     -1,     -1,     -1,     -1, 
    0x0640, 0x0641, 0x0642, 0x0643, 0x0644, 0x0645, 0x0646, 0x0647, 
    0x0648, 0x0649, 0x064a, 0x064b, 0x064c, 0x064d, 0x064e, 0x064f, 
    0x0650, 0x0651, 0x0652,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1
};

static int iso8859_7[224] = {
    0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027, 
    0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f, 
    0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037, 
    0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f, 
    0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047, 
    0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f, 
    0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057, 
    0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x005e, 0x005f, 
    0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067, 
    0x0068, 0x0069, 0x006a, 0x006b, 0x006c, 0x006d, 0x006e, 0x006f, 
    0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077, 
    0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d, 0x007e,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    0x00a0, 0x02bd, 0x02bc, 0x00a3,     -1,     -1, 0x00a6, 0x00a7, 
    0x00a8, 0x00a9,     -1, 0x00ab, 0x00ac, 0x00ad,     -1, 0x2015, 
    0x00b0, 0x00b1, 0x00b2, 0x00b3, 0x0384, 0x0385, 0x0386, 0x00b7, 
    0x0388, 0x0389, 0x038a, 0x00bb, 0x038c, 0x00bd, 0x038e, 0x038f, 
    0x0390, 0x0391, 0x0392, 0x0393, 0x0394, 0x0395, 0x0396, 0x0397, 
    0x0398, 0x0399, 0x039a, 0x039b, 0x039c, 0x039d, 0x039e, 0x039f, 
    0x03a0, 0x03a1,     -1, 0x03a3, 0x03a4, 0x03a5, 0x03a6, 0x03a7, 
    0x03a8, 0x03a9, 0x03aa, 0x03ab, 0x03ac, 0x03ad, 0x03ae, 0x03af, 
    0x03b0, 0x03b1, 0x03b2, 0x03b3, 0x03b4, 0x03b5, 0x03b6, 0x03b7, 
    0x03b8, 0x03b9, 0x03ba, 0x03bb, 0x03bc, 0x03bd, 0x03be, 0x03bf, 
    0x03c0, 0x03c1, 0x03c2, 0x03c3, 0x03c4, 0x03c5, 0x03c6, 0x03c7, 
    0x03c8, 0x03c9, 0x03ca, 0x03cb, 0x03cc, 0x03cd, 0x03ce,     -1
};

static int iso8859_8[224] = {
    0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027, 
    0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f, 
    0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037, 
    0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f, 
    0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047, 
    0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f, 
    0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057, 
    0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x005e, 0x005f, 
    0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067, 
    0x0068, 0x0069, 0x006a, 0x006b, 0x006c, 0x006d, 0x006e, 0x006f, 
    0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077, 
    0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d, 0x007e,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    0x00a0,     -1, 0x00a2, 0x00a3, 0x00a4, 0x00a5, 0x00a6, 0x00a7, 
    0x00a8, 0x00a9, 0x00d7, 0x00ab, 0x00ac, 0x00ad, 0x00ae, 0x203e, 
    0x00b0, 0x00b1, 0x00b2, 0x00b3, 0x00b4, 0x00b5, 0x00b6, 0x00b7, 
    0x00b8, 0x00b9, 0x00f7, 0x00bb, 0x00bc, 0x00bd, 0x00be,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1, 0x2017, 
    0x05d0, 0x05d1, 0x05d2, 0x05d3, 0x05d4, 0x05d5, 0x05d6, 0x05d7, 
    0x05d8, 0x05d9, 0x05da, 0x05db, 0x05dc, 0x05dd, 0x05de, 0x05df, 
    0x05e0, 0x05e1, 0x05e2, 0x05e3, 0x05e4, 0x05e5, 0x05e6, 0x05e7, 
    0x05e8, 0x05e9, 0x05ea,     -1,     -1,     -1,     -1,     -1 
};

static int iso8859_9[224] = {
    0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027, 
    0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f, 
    0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037, 
    0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f, 
    0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047, 
    0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f, 
    0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057, 
    0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x005e, 0x005f, 
    0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067, 
    0x0068, 0x0069, 0x006a, 0x006b, 0x006c, 0x006d, 0x006e, 0x006f, 
    0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077, 
    0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d, 0x007e,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    0x00a0, 0x00a1, 0x00a2, 0x00a3, 0x00a4, 0x00a5, 0x00a6, 0x00a7, 
    0x00a8, 0x00a9, 0x00aa, 0x00ab, 0x00ac, 0x00ad, 0x00ae, 0x00af, 
    0x00b0, 0x00b1, 0x00b2, 0x00b3, 0x00b4, 0x00b5, 0x00b6, 0x00b7, 
    0x00b8, 0x00b9, 0x00ba, 0x00bb, 0x00bc, 0x00bd, 0x00be, 0x00bf, 
    0x00c0, 0x00c1, 0x00c2, 0x00c3, 0x00c4, 0x00c5, 0x00c6, 0x00c7, 
    0x00c8, 0x00c9, 0x00ca, 0x00cb, 0x00cc, 0x00cd, 0x00ce, 0x00cf, 
    0x011e, 0x00d1, 0x00d2, 0x00d3, 0x00d4, 0x00d5, 0x00d6, 0x00d7, 
    0x00d8, 0x00d9, 0x00da, 0x00db, 0x00dc, 0x0130, 0x015e, 0x00df, 
    0x00e0, 0x00e1, 0x00e2, 0x00e3, 0x00e4, 0x00e5, 0x00e6, 0x00e7, 
    0x00e8, 0x00e9, 0x00ea, 0x00eb, 0x00ec, 0x00ed, 0x00ee, 0x00ef, 
    0x011f, 0x00f1, 0x00f2, 0x00f3, 0x00f4, 0x00f5, 0x00f6, 0x00f7, 
    0x00f8, 0x00f9, 0x00fa, 0x00fb, 0x00fc, 0x0131, 0x015f, 0x00ff 
};

static int iso8859_10[224] = {
    0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027,
    0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f,
    0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037,
    0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f,
    0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047,
    0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f,
    0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057,
    0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x005e, 0x005f,
    0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067,
    0x0068, 0x0069, 0x006a, 0x006b, 0x006c, 0x006d, 0x006e, 0x006f,
    0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077,
    0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d, 0x007e,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    0x00a0, 0x0104, 0x0112, 0x0122, 0x012a, 0x0128, 0x0136, 0x00a7,
    0x013b, 0x0110, 0x0160, 0x0166, 0x017d, 0x00ad, 0x016a, 0x014a,
    0x00b0, 0x0105, 0x0113, 0x0123, 0x012b, 0x0129, 0x0137, 0x00b7,
    0x013c, 0x0111, 0x0161, 0x0167, 0x017e, 0x2014, 0x016b, 0x014b,
    0x0100, 0x00c1, 0x00c2, 0x00c3, 0x00c4, 0x00c5, 0x00c6, 0x012e,
    0x010c, 0x00c9, 0x0118, 0x00cb, 0x0116, 0x00cd, 0x00ce, 0x00cf,
    0x00d0, 0x0145, 0x014c, 0x00d3, 0x00d4, 0x00d5, 0x00d6, 0x0168,
    0x00d8, 0x0172, 0x00da, 0x00db, 0x00dc, 0x00dd, 0x00de, 0x00df,
    0x0101, 0x00e1, 0x00e2, 0x00e3, 0x00e4, 0x00e5, 0x00e6, 0x012f,
    0x010d, 0x00e9, 0x0119, 0x00eb, 0x0117, 0x00ed, 0x00ee, 0x00ef,
    0x011f, 0x0146, 0x014d, 0x00f3, 0x00f4, 0x00f5, 0x00f6, 0x0169,
    0x00f8, 0x0173, 0x00fa, 0x00fb, 0x00fc, 0x00fd, 0x00fe, 0x0138
};

static int iso8859_11[224] = { /* Thai */
    0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027,
    0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f,
    0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037,
    0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f,
    0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047,
    0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f,
    0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057,
    0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x005e, 0x005f,
    0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067,
    0x0068, 0x0069, 0x006a, 0x006b, 0x006c, 0x006d, 0x006e, 0x006f,
    0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077,
    0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d, 0x007e,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    0x00a0, 0x0e01, 0x0e02, 0x0e03, 0x0e04, 0x0e05, 0x0e06, 0x0e07,
    0x0e08, 0x0e09, 0x0e0a, 0x0e0b, 0x0e0c, 0x0e0d, 0x0e0e, 0x0e0f,
    0x0e10, 0x0e11, 0x0e12, 0x0e13, 0x0e14, 0x0e15, 0x0e16, 0x0e17,
    0x0e18, 0x0019, 0x0e1a, 0x0e1b, 0x0e1c, 0x0e1d, 0x0e1e, 0x0e1f,
    0x0e20, 0x0e21, 0x0e22, 0x0e23, 0x0e24, 0x0e25, 0x0e26, 0x0e27,
    0x0e28, 0x0e29, 0x0e2a, 0x0e2b, 0x0e2c, 0x0e2d, 0x0e2e, 0x0e2f,
    0x0e30, 0x0e31, 0x0e32, 0x0e33, 0x0e34, 0x0e35, 0x0e36, 0x0e37,
    0x0e38, 0x0e39, 0x0e3a,     -1,     -1,     -1,     -1, 0x0e3f,
    0x0e40, 0x0e41, 0x0e42, 0x0e43, 0x0e44, 0x0e45, 0x0e46, 0x0e47,
    0x0e48, 0x0e49, 0x0e4a, 0x0e4b, 0x0e4c, 0x0e4d, 0x0e4e, 0x0e4f,
    0x0e50, 0x0e50, 0x0e50, 0x0e50, 0x0e50, 0x0e50, 0x0e50, 0x0e50,
    0x0e50, 0x0e50,     -1,     -1,     -1,     -1,     -1,     -1
};

static int tis620_2[224] = { /* Thai - Win extensions */
    0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027,
    0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f,
    0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037,
    0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f,
    0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047,
    0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f,
    0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057,
    0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x005e, 0x005f,
    0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067,
    0x0068, 0x0069, 0x006a, 0x006b, 0x006c, 0x006d, 0x006e, 0x006f,
    0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077,
    0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d, 0x007e,     -1,
    0xf700, 0xf701, 0xf702, 0xf703, 0xf704, 0x2026, 0xf705, 0xf706,
    0xf707, 0xf708, 0xf709, 0xf70a, 0xf70b, 0xf70c, 0xf70d, 0xf70e,
    0xf70f, 0x2018, 0x2019, 0x201c, 0x201d, 0x2022, 0x2013, 0x2014,
    0xf710, 0xf711, 0xf712, 0xf713, 0xf714, 0xf715, 0xf716, 0xf717,
    0x00a0, 0x0e01, 0x0e02, 0x0e03, 0x0e04, 0x0e05, 0x0e06, 0x0e07,
    0x0e08, 0x0e09, 0x0e0a, 0x0e0b, 0x0e0c, 0x0e0d, 0x0e0e, 0x0e0f,
    0x0e10, 0x0e11, 0x0e12, 0x0e13, 0x0e14, 0x0e15, 0x0e16, 0x0e17,
    0x0e18, 0x0019, 0x0e1a, 0x0e1b, 0x0e1c, 0x0e1d, 0x0e1e, 0x0e1f,
    0x0e20, 0x0e21, 0x0e22, 0x0e23, 0x0e24, 0x0e25, 0x0e26, 0x0e27,
    0x0e28, 0x0e29, 0x0e2a, 0x0e2b, 0x0e2c, 0x0e2d, 0x0e2e, 0x0e2f,
    0x0e30, 0x0e31, 0x0e32, 0x0e33, 0x0e34, 0x0e35, 0x0e36, 0x0e37,
    0x0e38, 0x0e39, 0x0e3a,     -1,     -1,     -1,     -1, 0x0e3f,
	0x0e40, 0x0e41, 0x0e42, 0x0e43, 0x0e44, 0x0e45, 0x0e46, 0x0e47,
    0x0e48, 0x0e49, 0x0e4a, 0x0e4b, 0x0e4c, 0x0e4d, 0x0e4e, 0x0e4f,
    0x0e50, 0x0e51, 0x0e52, 0x0e53, 0x0e54, 0x0e55, 0x0e56, 0x0e57,
    0x0e58, 0x0e59, 0x0e5a, 0x0e5b, 0xf718, 0xf719, 0xf71a,     -1
};

static int iso8859_13[224] = {
    0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027,
    0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f,
    0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037,
    0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f,
    0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047,
    0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f,
    0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057,
    0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x005e, 0x005f,
    0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067,
    0x0068, 0x0069, 0x006a, 0x006b, 0x006c, 0x006d, 0x006e, 0x006f,
    0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077,
    0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d, 0x007e,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    0x00a0, 0x201d, 0x00a2, 0x00a3, 0x00a4, 0x201e, 0x00a6, 0x00a7,
    0x00d8, 0x00a9, 0x0156, 0x00ab, 0x00ac, 0x00ad, 0x00ae, 0x00c6,
    0x00b0, 0x00b1, 0x00b2, 0x00b3, 0x201c, 0x00b5, 0x00b6, 0x00b7,
    0x00f8, 0x00b9, 0x0157, 0x00bb, 0x00bc, 0x00bd, 0x00be, 0x00e6,
    0x0104, 0x012e, 0x0100, 0x0106, 0x00c4, 0x00c5, 0x0118, 0x0112,
    0x010c, 0x00c9, 0x0179, 0x0116, 0x0122, 0x0136, 0x012a, 0x013b,
    0x0160, 0x0143, 0x0145, 0x00d3, 0x014c, 0x00d5, 0x00d6, 0x00d7,
    0x0172, 0x0141, 0x015a, 0x016a, 0x00dc, 0x017b, 0x017d, 0x00df,
    0x0105, 0x012f, 0x0101, 0x0107, 0x00e4, 0x00e5, 0x0119, 0x0113,
    0x010d, 0x00e9, 0x017a, 0x0117, 0x0123, 0x0137, 0x012b, 0x013c,
    0x0161, 0x0144, 0x0146, 0x00f3, 0x014d, 0x00f5, 0x00f6, 0x00f7,
    0x0173, 0x0142, 0x015b, 0x016b, 0x00fc, 0x017c, 0x017e, 0x2019
};

static int iso8859_14[224] = {
    0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027,
    0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f,
    0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037,
    0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f,
    0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047,
    0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f,
    0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057,
    0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x005e, 0x005f,
    0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067,
    0x0068, 0x0069, 0x006a, 0x006b, 0x006c, 0x006d, 0x006e, 0x006f,
    0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077,
    0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d, 0x007e,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
/*  0x00a0, 0x1e02, 0x1e03, 0x00a3, 0x010a, 0x010b, 0x1e0a, 0x00a7,
    0x1e80, 0x00a9, 0x1e82, 0x1e0b, 0x1ef2, 0x00ad, 0x00ae, 0x0178,
    0x1e1e, 0x1e1f, 0x0120, 0x0121, 0x1e40, 0x1e41, 0x00b6, 0x1e56,
    0x1e81, 0x1e57, 0x1e83, 0x1e60, 0x1ef3, 0x1e84, 0x1e85, 0x1e61, */
    0x00a0,     -1,     -1, 0x00a3, 0x010a, 0x010b,     -1, 0x00a7,
    0x1e80, 0x00a9, 0x1e82,     -1, 0x1ef2, 0x00ad, 0x00ae, 0x0178,
        -1, 0x1e57, 0x1e83,     -1,     -1,     -1, 0x00b6,     -1,
    0x1e81,     -1, 0x1e83,     -1, 0x1ef3, 0x1e84, 0x1e85,     -1,
    0x00c0, 0x00c1, 0x00c2, 0x00c3, 0x00c4, 0x00c5, 0x00c6, 0x00c7,
    0x00c8, 0x00c9, 0x00ca, 0x00cb, 0x00cc, 0x00cd, 0x00ce, 0x00cf,
    0x0174, 0x00d1, 0x00d2, 0x00d3, 0x00d4, 0x00d5, 0x00d6, 0x1e6a,
    0x00d8, 0x00d9, 0x00da, 0x00db, 0x00dc, 0x00dd, 0x0176, 0x00df,
    0x00e0, 0x00e1, 0x00e2, 0x00e3, 0x00e4, 0x00e5, 0x00e6, 0x00e7,
    0x00e8, 0x00e9, 0x00ea, 0x00eb, 0x00ec, 0x00ed, 0x00ee, 0x00ef,
    0x0175, 0x00f1, 0x00f2, 0x00f3, 0x00f4, 0x00f5, 0x00f6, 0x1e6b,
    0x00f8, 0x00f9, 0x00fa, 0x00fb, 0x00fc, 0x00fd, 0x0177, 0x00ff
};

static int iso8859_15[224] = {
    0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027,
    0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f,
    0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037,
    0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f,
    0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047,
    0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f,
    0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057,
    0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x005e, 0x005f,
    0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067,
    0x0068, 0x0069, 0x006a, 0x006b, 0x006c, 0x006d, 0x006e, 0x006f,
    0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077,
    0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d, 0x007e,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    0x00a0, 0x00a1, 0x00a2, 0x00a3, 0x20ac, 0x00a5, 0x0160, 0x00a7,
    0x0161, 0x00a9, 0x00aa, 0x00ab, 0x00ac, 0x00ad, 0x00ae, 0x00af,
    0x00b0, 0x00b1, 0x00b2, 0x00b3, 0x017d, 0x00b5, 0x00b6, 0x00b7,
    0x017e, 0x00b9, 0x00ba, 0x00bb, 0x0152, 0x0153, 0x0178, 0x00bf,
    0x00c0, 0x00c1, 0x00c2, 0x00c3, 0x00c4, 0x00c5, 0x00c6, 0x00c7,
    0x00c8, 0x00c9, 0x00ca, 0x00cb, 0x00cc, 0x00cd, 0x00ce, 0x00cf,
    0x00d0, 0x00d1, 0x00d2, 0x00d3, 0x00d4, 0x00d5, 0x00d6, 0x00d7,
    0x00d8, 0x00d9, 0x00da, 0x00db, 0x00dc, 0x00dd, 0x00de, 0x00df,
    0x00e0, 0x00e1, 0x00e2, 0x00e3, 0x00e4, 0x00e5, 0x00e6, 0x00e7,
    0x00e8, 0x00e9, 0x00ea, 0x00eb, 0x00ec, 0x00ed, 0x00ee, 0x00ef,
    0x00f0, 0x00f1, 0x00f2, 0x00f3, 0x00f4, 0x00f5, 0x00f6, 0x00f7,
    0x00f8, 0x00f9, 0x00fa, 0x00fb, 0x00fc, 0x00fd, 0x00fe, 0x00ff
};

static int iso8859_16[224] = {
    0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027, 
    0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f, 
    0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037, 
    0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f, 
    0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047, 
    0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f, 
    0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057, 
    0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x005e, 0x005f, 
    0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067, 
    0x0068, 0x0069, 0x006a, 0x006b, 0x006c, 0x006d, 0x006e, 0x006f, 
    0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077, 
    0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d, 0x007e,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    0x00a0, 0x0104, 0x0105, 0x0141, 0x20ac, 0x00ab, 0x0160, 0x00a7, 
    0x0161, 0x00a9, 0x0218, 0x201e, 0x0179, 0x00ad, 0x017a, 0x017b, 
    0x00b0, 0x00b1, 0x010c, 0x0142, 0x017d, 0x201d, 0x00b6, 0x00b7, 
    0x017e, 0x010d, 0x0219, 0x00bb, 0x0152, 0x0153, 0x0178, 0x017c, 
    0x00c0, 0x00c1, 0x00c2, 0x0102, 0x00c4, 0x0106, 0x00c6, 0x00c7, 
    0x00c8, 0x00c9, 0x00ca, 0x00cb, 0x00cc, 0x00cd, 0x00ce, 0x00cf, 
    0x0110, 0x0143, 0x00d2, 0x00d3, 0x00d4, 0x0150, 0x00d6, 0x015a, 
    0x0170, 0x00d9, 0x00da, 0x00db, 0x00dc, 0x0118, 0x021a, 0x00df, 
    0x00e0, 0x00e1, 0x00e2, 0x0103, 0x00e4, 0x0107, 0x00e6, 0x00e7, 
    0x00e8, 0x00e9, 0x00ea, 0x00eb, 0x00ec, 0x00ed, 0x00ee, 0x00ef, 
    0x0111, 0x0144, 0x00f2, 0x00f3, 0x00f4, 0x0151, 0x00f6, 0x015b, 
    0x0171, 0x00f9, 0x00fa, 0x00fb, 0x00fc, 0x0119, 0x021b, 0x00ff 
};

static int iso8859_9e[224] = {
    0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027, 
    0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f, 
    0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037, 
    0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f, 
    0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047, 
    0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f, 
    0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057, 
    0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x005e, 0x005f, 
    0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067, 
    0x0068, 0x0069, 0x006a, 0x006b, 0x006c, 0x006d, 0x006e, 0x006f, 
    0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077, 
    0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d, 0x007e,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    0x00a0, 0x017d, 0x00a2, 0x00a3, 0x20ac, 0x00a5, 0x012c, 0x00a7, 
    0x016c, 0x00a9, 0x01e6, 0x00ab, 0x014a, 0x00ad, 0x00ae, 0x01d1, 
    0x00b0, 0x017e, 0x00b2, 0x00b3, 0x00b4, 0x00b5, 0x012d, 0x00b7, 
    0x016d, 0x00b9, 0x01e7, 0x00bb, 0x014b, 0x00bd, 0x0178, 0x01d2, 
    0x00c0, 0x00c1, 0x00c2, 0x00c3, 0x00c4, 0x00c5, 0x018f, 0x00c7, 
    0x00c8, 0x00c9, 0x00ca, 0x00cb, 0x00cc, 0x00cd, 0x00ce, 0x00cf, 
    0x011e, 0x00d1, 0x00d2, 0x00d3, 0x00d4, 0x00d5, 0x00d6, 0x00dd, 
    0x019f, 0x00d9, 0x00da, 0x00db, 0x00dc, 0x0130, 0x015e, 0x00df, 
    0x00e0, 0x00e1, 0x00e2, 0x00e3, 0x00e4, 0x00e5, 0x0259, 0x00e7, 
    0x00e8, 0x00e9, 0x00ea, 0x00eb, 0x00ec, 0x00ed, 0x00ee, 0x00ef, 
    0x011f, 0x00f1, 0x00f2, 0x00f3, 0x00f4, 0x00f5, 0x00f6, 0x00fd, 
    0x0275, 0x00f9, 0x00fa, 0x00fb, 0x00fc, 0x0131, 0x015f, 0x00ff 
};

static int koi8r[224] = {
    0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027, 
    0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f, 
    0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037, 
    0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f, 
    0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047, 
    0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f, 
    0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057, 
    0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x005e, 0x005f, 
    0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067, 
    0x0068, 0x0069, 0x006a, 0x006b, 0x006c, 0x006d, 0x006e, 0x006f, 
    0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077, 
    0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d, 0x007e,     -1, 
    0x2500, 0x2502, 0x250c, 0x2510, 0x2514, 0x2518, 0x251c, 0x2524, 
    0x252c, 0x2534, 0x253c, 0x2580, 0x2584, 0x2588, 0x258c, 0x2590, 
    0x2591, 0x2592, 0x2593, 0x2320, 0x2043, 0x2022, 0x221a, 0x2248,
    0x2264, 0x2265, 0x00a0, 0x2321, 0x00b0, 0x00b2, 0x00b7, 0x00f7, 
    0x2550, 0x2551, 0x2552, 0x0451, 0x2553, 0x2554, 0x2555, 0x2556, 
    0x2557, 0x2558, 0x2559, 0x255a, 0x255b, 0x255c, 0x255d, 0x255e, 
    0x255f, 0x2560, 0x2561, 0x0401, 0x2562, 0x2563, 0x2564, 0x2565, 
    0x2566, 0x2567, 0x2568, 0x2569, 0x256a, 0x256b, 0x256c, 0x00a9, 
    0x044e, 0x0430, 0x0431, 0x0446, 0x0434, 0x0435, 0x0444, 0x0433, 
    0x0445, 0x0438, 0x0439, 0x043a, 0x043b, 0x043c, 0x043d, 0x043e, 
    0x043f, 0x044f, 0x0440, 0x0441, 0x0442, 0x0443, 0x0436, 0x0432, 
    0x044c, 0x044b, 0x0437, 0x0448, 0x044d, 0x0449, 0x0447, 0x044a, 
    0x042e, 0x0410, 0x0411, 0x0426, 0x0414, 0x0415, 0x0424, 0x0413, 
    0x0425, 0x0418, 0x0419, 0x041a, 0x041b, 0x041c, 0x041d, 0x041e, 
    0x041f, 0x042f, 0x0420, 0x0421, 0x0422, 0x0423, 0x0416, 0x0412, 
    0x042c, 0x042b, 0x0417, 0x0428, 0x042d, 0x0429, 0x0427, 0x042a 
};

static int koi8ru[224] = {
    0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027, 
    0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f, 
    0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037, 
    0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f, 
    0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047, 
    0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f, 
    0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057, 
    0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x005e, 0x005f, 
    0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067, 
    0x0068, 0x0069, 0x006a, 0x006b, 0x006c, 0x006d, 0x006e, 0x006f, 
    0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077, 
    0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d, 0x007e,     -1, 
    0x2500, 0x2502, 0x250c, 0x2510, 0x2514, 0x2518, 0x251c, 0x2524, 
    0x252c, 0x2534, 0x253c, 0x2580, 0x2584, 0x2588, 0x258c, 0x2590, 
    0x2591, 0x2592, 0x2593, 0x201c, 0x2043, 0x2022, 0x201d, 0x2014,
    0x2116, 0x2122, 0x00a0, 0x00bb, 0x00ae, 0x00ab, 0x00b7, 0x00a4, 
    0x2550, 0x2551, 0x2552, 0x0451, 0x0454, 0x2554, 0x0456, 0x0457, 
    0x2557, 0x2558, 0x2559, 0x255a, 0x255b, 0x0491, 0x045e, 0x255e, 
    0x255f, 0x2560, 0x2561, 0x0401, 0x0404, 0x2563, 0x0406, 0x0407, 
    0x2566, 0x2567, 0x2568, 0x2569, 0x256a, 0x0490, 0x040e, 0x00a9, 
    0x044e, 0x0430, 0x0431, 0x0446, 0x0434, 0x0435, 0x0444, 0x0433, 
    0x0445, 0x0438, 0x0439, 0x043a, 0x043b, 0x043c, 0x043d, 0x043e, 
    0x043f, 0x044f, 0x0440, 0x0441, 0x0442, 0x0443, 0x0436, 0x0432, 
    0x044c, 0x044b, 0x0437, 0x0448, 0x044d, 0x0449, 0x0447, 0x044a, 
    0x042e, 0x0410, 0x0411, 0x0426, 0x0414, 0x0415, 0x0424, 0x0413, 
    0x0425, 0x0418, 0x0419, 0x041a, 0x041b, 0x041c, 0x041d, 0x041e, 
    0x041f, 0x042f, 0x0420, 0x0421, 0x0422, 0x0423, 0x0416, 0x0412, 
    0x042c, 0x042b, 0x0417, 0x0428, 0x042d, 0x0429, 0x0427, 0x042a 
};

static int koi8r1[224] = {
    0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027, 
    0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f, 
    0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037, 
    0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f, 
    0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047, 
    0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f, 
    0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057, 
    0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x005e, 0x005f, 
    0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067, 
    0x0068, 0x0069, 0x006a, 0x006b, 0x006c, 0x006d, 0x006e, 0x006f, 
    0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077, 
    0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d, 0x007e,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1, 0x2320, 0x2043, 0x2022, 0x221a, 0x2248, 
    0x2264, 0x2265, 0x00a0, 0x2321, 0x00b0, 0x00b2, 0x00b7, 0x00f7, 
    /**/-1,     -1,     -1, 0x0451,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1, 0x0401,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1, 0x00a9, 
    0x044e, 0x0430, 0x0431, 0x0446, 0x0434, 0x0435, 0x0444, 0x0433, 
    0x0445, 0x0438, 0x0439, 0x043a, 0x043b, 0x043c, 0x043d, 0x043e, 
    0x043f, 0x044f, 0x0440, 0x0441, 0x0442, 0x0443, 0x0436, 0x0432, 
    0x044c, 0x044b, 0x0437, 0x0448, 0x044d, 0x0449, 0x0447, 0x044a, 
    0x042e, 0x0410, 0x0411, 0x0426, 0x0414, 0x0415, 0x0424, 0x0413, 
    0x0425, 0x0418, 0x0419, 0x041a, 0x041b, 0x041c, 0x041d, 0x041e, 
    0x041f, 0x042f, 0x0420, 0x0421, 0x0422, 0x0423, 0x0416, 0x0412, 
    0x042c, 0x042b, 0x0417, 0x0428, 0x042d, 0x0429, 0x0427, 0x042a, 
};

static int koi8ru1[224] = {
    0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027, 
    0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f, 
    0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037, 
    0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f, 
    0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047, 
    0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f, 
    0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057, 
    0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x005e, 0x005f, 
    0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067, 
    0x0068, 0x0069, 0x006a, 0x006b, 0x006c, 0x006d, 0x006e, 0x006f, 
    0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077, 
    0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d, 0x007e,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1, 0x201c, 0x2043, 0x2022, 0x201d, 0x2014, 
    0x2116, 0x2122, 0x00a0, 0x00bb, 0x00ae, 0x00ab, 0x00b7, 0x00a4, 
    /**/-1,     -1,     -1, 0x0451, 0x0454,     -1, 0x0456, 0x0457, 
    /**/-1,     -1,     -1,     -1,     -1, 0x0491, 0x045e,     -1, 
    /**/-1,     -1,     -1, 0x0401, 0x0404,     -1, 0x0406, 0x0407, 
    /**/-1,     -1,     -1,     -1,     -1, 0x0490, 0x040e, 0x00a9, 
    0x044e, 0x0430, 0x0431, 0x0446, 0x0434, 0x0435, 0x0444, 0x0433, 
    0x0445, 0x0438, 0x0439, 0x043a, 0x043b, 0x043c, 0x043d, 0x043e, 
    0x043f, 0x044f, 0x0440, 0x0441, 0x0442, 0x0443, 0x0436, 0x0432, 
    0x044c, 0x044b, 0x0437, 0x0448, 0x044d, 0x0449, 0x0447, 0x044a, 
    0x042e, 0x0410, 0x0411, 0x0426, 0x0414, 0x0415, 0x0424, 0x0413, 
    0x0425, 0x0418, 0x0419, 0x041a, 0x041b, 0x041c, 0x041d, 0x041e, 
    0x041f, 0x042f, 0x0420, 0x0421, 0x0422, 0x0423, 0x0416, 0x0412, 
    0x042c, 0x042b, 0x0417, 0x0428, 0x042d, 0x0429, 0x0427, 0x042a, 
};

static int koi8k[224] = {
    0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027, 
    0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f, 
    0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037, 
    0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f, 
    0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047, 
    0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f, 
    0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057, 
    0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x005e, 0x005f, 
    0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067, 
    0x0068, 0x0069, 0x006a, 0x006b, 0x006c, 0x006d, 0x006e, 0x006f, 
    0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077, 
    0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d, 0x007e,     -1, 
    0x0493, 0x0497, 0x049b, 0x049d, 0x04a3, 0x04af, 0x04b1, 0x04b3, 
    0x04b7, 0x04b9, 0x04bb,     -1, 0x04d9, 0x04e3, 0x04e9, 0x04ef,
    0x0492, 0x0496, 0x049a, 0x049c, 0x04a2, 0x04ae, 0x04b0, 0x04b2, 
    0x04b6, 0x04b8, 0x04ba,     -1, 0x04d8, 0x04e2, 0x04e8, 0x04ee, 
        -1, 0x0452, 0x0453, 0x0451, 0x0454, 0x0455, 0x0456, 0x0457, 
    0x0458, 0x0459, 0x045a, 0x045b, 0x045c, 0x0491, 0x045e, 0x045f, 
        -1, 0x0402, 0x0403, 0x0401, 0x0404, 0x0405, 0x0406, 0x0407, 
    0x0408, 0x0409, 0x040a, 0x040b, 0x040c, 0x0490, 0x040e, 0x040f, 
    0x044e, 0x0430, 0x0431, 0x0446, 0x0434, 0x0435, 0x0444, 0x0433, 
    0x0445, 0x0438, 0x0439, 0x043a, 0x043b, 0x043c, 0x043d, 0x043e, 
    0x043f, 0x044f, 0x0440, 0x0441, 0x0442, 0x0443, 0x0436, 0x0432, 
    0x044c, 0x044b, 0x0437, 0x0448, 0x044d, 0x0449, 0x0447, 0x044a, 
    0x042e, 0x0410, 0x0411, 0x0426, 0x0414, 0x0415, 0x0424, 0x0413, 
    0x0425, 0x0418, 0x0419, 0x041a, 0x041b, 0x041c, 0x041d, 0x041e, 
    0x041f, 0x042f, 0x0420, 0x0421, 0x0422, 0x0423, 0x0416, 0x0412, 
    0x042c, 0x042b, 0x0417, 0x0428, 0x042d, 0x0429, 0x0427, 0x042a, 
};

static int koi8t[224] = {
    0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027, 
    0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f, 
    0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037, 
    0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f, 
    0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047, 
    0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f, 
    0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057, 
    0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x005e, 0x005f, 
    0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067, 
    0x0068, 0x0069, 0x006a, 0x006b, 0x006c, 0x006d, 0x006e, 0x006f, 
    0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077, 
    0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d, 0x007e,     -1, 
    0x049b, 0x0493, 0x201a, 0x0492, 0x201e, 0x2026, 0x2020, 0x2021, 
        -1, 0x2030, 0x04b3, 0x2039, 0x04b2, 0x04b7, 0x04b6,     -1,
    0x049a, 0x2018, 0x2019, 0x201c, 0x201d, 0x2022, 0x2013, 0x2014, 
        -1, 0x2122,     -1, 0x203a,     -1,     -1,     -1,     -1, 
        -1, 0x04ef, 0x04ee, 0x0451, 0x00a4, 0x04e3, 0x00a6, 0x00a7, 
        -1,     -1,     -1, 0x00ab, 0x00ac, 0x00ad, 0x00ae,     -1, 
    0x00b0, 0x00b1, 0x00b2, 0x0401,     -1, 0x04e2, 0x00b6, 0x00b7, 
        -1, 0x2116,     -1, 0x00bb,     -1,     -1,     -1, 0x00a9, 
    0x044e, 0x0430, 0x0431, 0x0446, 0x0434, 0x0435, 0x0444, 0x0433, 
    0x0445, 0x0438, 0x0439, 0x043a, 0x043b, 0x043c, 0x043d, 0x043e, 
    0x043f, 0x044f, 0x0440, 0x0441, 0x0442, 0x0443, 0x0436, 0x0432, 
    0x044c, 0x044b, 0x0437, 0x0448, 0x044d, 0x0449, 0x0447, 0x044a, 
    0x042e, 0x0410, 0x0411, 0x0426, 0x0414, 0x0415, 0x0424, 0x0413, 
    0x0425, 0x0418, 0x0419, 0x041a, 0x041b, 0x041c, 0x041d, 0x041e, 
    0x041f, 0x042f, 0x0420, 0x0421, 0x0422, 0x0423, 0x0416, 0x0412, 
    0x042c, 0x042b, 0x0417, 0x0428, 0x042d, 0x0429, 0x0427, 0x042a, 
};

static int mscp1251[224] = {
     0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027,
     0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f,
     0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037,
     0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f,
     0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047,
     0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f,
     0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057,
     0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x005e, 0x005f,
     0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067,
     0x0068, 0x0069, 0x006a, 0x006b, 0x006c, 0x006d, 0x006e, 0x006f,
     0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077,
     0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d, 0x007e, 0x007f,
     0x0402, 0x0403, 0x201a, 0x0453, 0x201e, 0x2026, 0x2020, 0x2021,
     0x20ac, 0x2030, 0x0409, 0x2039, 0x040a, 0x040c, 0x040b, 0x040f,
     0x0452, 0x2018, 0x2019, 0x201c, 0x201d, 0x2022, 0x2013, 0x2014,
         -1, 0x2122, 0x0459, 0x203a, 0x045a, 0x045c, 0x045b, 0x045f,
     0x00a0, 0x040e, 0x045e, 0x0408, 0x00a4, 0x0490, 0x00a6, 0x00a7,
     0x0401, 0x00a9, 0x0404, 0x00ab, 0x00ac, 0x00ad, 0x00ae, 0x0407,
     0x00b0, 0x00b1, 0x0406, 0x0456, 0x0491, 0x00b5, 0x00b6, 0x00b7,
     0x0451, 0x2116, 0x0454, 0x00bb, 0x0458, 0x0405, 0x0455, 0x0457,
     0x0410, 0x0411, 0x0412, 0x0413, 0x0414, 0x0415, 0x0416, 0x0417,
     0x0418, 0x0419, 0x041a, 0x041b, 0x041c, 0x041d, 0x041e, 0x041f,
     0x0420, 0x0421, 0x0422, 0x0423, 0x0424, 0x0425, 0x0426, 0x0427,
     0x0428, 0x0429, 0x042a, 0x042b, 0x042c, 0x042d, 0x042e, 0x042f,
     0x0430, 0x0431, 0x0432, 0x0433, 0x0434, 0x0435, 0x0436, 0x0437,
     0x0438, 0x0439, 0x043a, 0x043b, 0x043c, 0x043d, 0x043e, 0x043f,
     0x0440, 0x0441, 0x0442, 0x0443, 0x0444, 0x0445, 0x0446, 0x0447,
     0x0448, 0x0449, 0x044a, 0x044b, 0x044c, 0x044d, 0x044e, 0x044f
};

static int mscp1251tt[224] = {
     0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027,
     0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f,
     0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037,
     0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f,
     0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047,
     0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f,
     0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057,
     0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x005e, 0x005f,
     0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067,
     0x0068, 0x0069, 0x006a, 0x006b, 0x006c, 0x006d, 0x006e, 0x006f,
     0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077,
     0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d, 0x007e, 0x007f,
     0x04d8, 0x0403, 0x201a, 0x0453, 0x201e, 0x2026, 0x2020, 0x2021,
     0x20ac, 0x2030, 0x04e8, 0x2039, 0x04ae, 0x0496, 0x04a2, 0x04ba,
     0x04d9, 0x2018, 0x2019, 0x201c, 0x201d, 0x2022, 0x2013, 0x2014,
         -1, 0x2122, 0x04e9, 0x203a, 0x04af, 0x0497, 0x04a3, 0x04bb,
     0x00a0, 0x040e, 0x045e, 0x0408, 0x00a4, 0x0490, 0x00a6, 0x00a7,
     0x0401, 0x00a9, 0x0404, 0x00ab, 0x00ac, 0x00ad, 0x00ae, 0x0407,
     0x00b0, 0x00b1, 0x0406, 0x0456, 0x0491, 0x00b5, 0x00b6, 0x00b7,
     0x0451, 0x2116, 0x0454, 0x00bb, 0x0458, 0x0405, 0x0455, 0x0457,
     0x0410, 0x0411, 0x0412, 0x0413, 0x0414, 0x0415, 0x0416, 0x0417,
     0x0418, 0x0419, 0x041a, 0x041b, 0x041c, 0x041d, 0x041e, 0x041f,
     0x0420, 0x0421, 0x0422, 0x0423, 0x0424, 0x0425, 0x0426, 0x0427,
     0x0428, 0x0429, 0x042a, 0x042b, 0x042c, 0x042d, 0x042e, 0x042f,
     0x0430, 0x0431, 0x0432, 0x0433, 0x0434, 0x0435, 0x0436, 0x0437,
     0x0438, 0x0439, 0x043a, 0x043b, 0x043c, 0x043d, 0x043e, 0x043f,
     0x0440, 0x0441, 0x0442, 0x0443, 0x0444, 0x0445, 0x0446, 0x0447,
     0x0448, 0x0449, 0x044a, 0x044b, 0x044c, 0x044d, 0x044e, 0x044f
};

static int mscp1255[224] = {
    0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027, 
    0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f, 
    0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037, 
    0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f, 
    0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047, 
    0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f, 
    0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057, 
    0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x005e, 0x005f, 
    0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067, 
    0x0068, 0x0069, 0x006a, 0x006b, 0x006c, 0x006d, 0x006e, 0x006f, 
    0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077, 
    0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d, 0x007e,     -1, 
    0x20ac,     -1, 0x201a, 0x0192, 0x201e, 0x2026, 0x2020, 0x2021,
    0x02c6, 0x2030,     -1, 0x2039,     -1,     -1,     -1,     -1,
        -1, 0x2018, 0x2019, 0x201c, 0x201d, 0x2022, 0x2013, 0x2014,
    0x02dc, 0x2122,     -1, 0x203a,     -1,     -1,     -1,     -1,
    0x00a0, 0x00a1, 0x00a2, 0x00a3, 0x20aa, 0x00a5, 0x00a6, 0x00a7, 
    0x00a8, 0x00a9, 0x00aa, 0x00ab, 0x00ac, 0x00ad, 0x00ae, 0x00af, 
    0x00b0, 0x00b1, 0x00b2, 0x00b3, 0x00b4, 0x00b5, 0x00b6, 0x00b7, 
    0x00b8, 0x00b9, 0x00f7, 0x00bb, 0x00bc, 0x00bd, 0x00be, 0x00bf, 
    0x05b0, 0x05b1, 0x05b2, 0x05b3, 0x05b4, 0x05b5, 0x05b6, 0x05b7, 
    0x05b8, 0x05b9,     -1, 0x05bb, 0x05bc, 0x05bd, 0x05be, 0x05bf, 
    0x05c0, 0x05c1, 0x05c2, 0x05c3, 0x05f0, 0x05f1, 0x05f2, 0x05f3, 
    0x05f4,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    0x05d0, 0x05d1, 0x05d2, 0x05d3, 0x05d4, 0x05d5, 0x05d6, 0x05d7, 
    0x05d8, 0x05d9, 0x05da, 0x05db, 0x05dc, 0x05dd, 0x05de, 0x05df, 
    0x05e0, 0x05e1, 0x05e2, 0x05e3, 0x05e4, 0x05e5, 0x05e6, 0x05e7, 
    0x05e8, 0x05e9, 0x05ea,     -1,     -1,     -1,     -1,     -1 
};

static int mscp1256[224] = {
     0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027,
     0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f,
     0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037,
     0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f,
     0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047,
     0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f,
     0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057,
     0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x005e, 0x005f,
     0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067,
     0x0068, 0x0069, 0x006a, 0x006b, 0x006c, 0x006d, 0x006e, 0x006f,
     0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077,
     0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d, 0x007e, 0x007f,
     0x20ac, 0x067e, 0x201a, 0x0192, 0x201e, 0x2026, 0x2020, 0x2021,
     0x02c6, 0x2030, 0x0679, 0x2039, 0x0152, 0x0686, 0x0698, 0x0688,
     0x06af, 0x2018, 0x2019, 0x201c, 0x201d, 0x2022, 0x2013, 0x2014,
     0x06a9, 0x2122, 0x0691, 0x203a, 0x0153,     -1,     -1, 0x06ba,
     0x00a0, 0x060c, 0x00a2, 0x00a3, 0x00a4, 0x00a5, 0x00a6, 0x00a7,
     0x00a8, 0x00a9, 0x06be, 0x00ab, 0x00ac, 0x00ad, 0x00ae, 0x00af,
     0x00b0, 0x00b1, 0x00b2, 0x00b3, 0x00b4, 0x00b5, 0x00b6, 0x00b7,
     0x00b8, 0x00b9, 0x061b, 0x00bb, 0x00bc, 0x00bd, 0x00be, 0x061f,
     0x06c1, 0x0621, 0x0622, 0x0623, 0x0624, 0x0625, 0x0626, 0x0627,
     0x0628, 0x0629, 0x062a, 0x062b, 0x062c, 0x062d, 0x062e, 0x062f,
     0x0630, 0x0631, 0x0632, 0x0633, 0x0634, 0x0635, 0x0636, 0x00d7,
     0x0637, 0x0638, 0x0639, 0x063a, 0x0640, 0x0641, 0x0642, 0x0643,
     0x00e0, 0x0644, 0x00e2, 0x0645, 0x0646, 0x0647, 0x0648, 0x00e7,
     0x00e8, 0x00e9, 0x00ea, 0x00eb, 0x0649, 0x064a, 0x00ee, 0x00ef,
     0x064b, 0x064c, 0x064d, 0x064e, 0x00f4, 0x064f, 0x0650, 0x00f7,
     0x0651, 0x00f9, 0x0652, 0x00fb, 0x00fc,     -1,     -1, 0x06d2
};

static int urdunaqsh0[224] = {
     0x0020, 0xfe81, 0xfe8d, 0xfe8f, 0xfb56, 0xfe95, 0xfb66, 0xfe99,
     0xfe9d, 0xfb7a, 0xfea1, 0xfea5, 0xfea9, 0xfb88, 0xfeab, 0xfead,
     0xfb8c, 0xfeaf, 0xfb8a, 0xfeb1, 0xfeb5, 0xfeb9, 0xfebd, 0xfec1,
     0xfec5, 0xfec9, 0xfecd, 0xfed1, 0xfb6a, 0xfed5, 0xfed9, 0xfb8e,
     0xfb92, 0xfedd, 0xfefb, 0xfee1, 0xfee5, 0xfb9e, 0xfeed, 0xfee9,
     0xfe93, 0xfbaa, 0xfe80, 0xfeef, 0xfef1, 0xfbae, 0xfe91, 0xfb58,
     0xfe97, 0xfb68, 0xfe9b, 0xfe9f, 0xfb7c, 0xfea3, 0xfea7, 0xfeb3,
     0xfeb7, 0xfebb, 0xfebf, 0xfec3, 0xfec7, 0xfecb, 0xfecf, 0xfed3,
     0xfb6c, 0xfed7, 0xfedb, 0xfb94, 0xfedf, 0xfee3, 0xfee7, 0xfba8,
     0xfeeb, 0xfe8b, 0xfef3, 0xfe92, 0xfb59, 0xfe98, 0xfb69, 0xfe96,
     0xfe9c, 0xfea0, 0xfb7d, 0xfea4, 0xfea8, 0xfeb4, 0xfbb8, 0xfebc,
     0xfec0, 0xfec4, 0xfec8, 0xfecc, 0xfed0,     -1,     -1,     -1,
         -1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
         -1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
         -1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
         -1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
         -1, 0xfed4, 0xfb6d, 0xfed8, 0xfedc, 0xfb95, 0xfee0, 0xfee4,
     0xfee9, 0xfba9, 0xfeec, 0xfe8c, 0xfef4, 0xfe8e, 0xfe90, 0xfb57,
     0xfe96, 0xfb67, 0xfe9a, 0xfe9e, 0xfb7b, 0xfea2, 0xfea6,     -1,
     0xfeaa, 0xfb89, 0xfeac, 0xfeae, 0xfb8d, 0xfeb2, 0xfeb6, 0xfeba,
     0xfebe, 0xfec2, 0xfec6, 0xfeca, 0xfece, 0xfed2, 0xfb6b, 0xfed6,
     0xfeda, 0xfb8f, 0xfb93, 0xfede, 0xfefc, 0xfee2, 0xfee6, 0xfb9f,
     0xfeee, 0xfba7, 0xfeea, 0xfe94, 0xfef0, 0xfef2, 0xfbaf, 0x0661,
     0x0662, 0x0663, 0x0664, 0x06f4, 0x0665, 0x06f5, 0x0666, 0x0667,
     0x06f7, 0x0668, 0x0669, 0x0660, 0x0640, 0x002e, 0x060c, 0x061f,
     0x0021, 0x003a, 0x0027, 0x0028, 0x0029,     -1,     -1,     -1,
         -1,     -1,     -1, 0x0653, 0x0670, 0x0655, 0x0654, 0x064c,
     0x064b, 0x064d,     -1,     -1, 0x0651, 0x064f, 0x064e, 0x0650
};

static int ethiopic[224] = { /* no need to check for them all... */
     0x1200, 0x1201, 0x1202, 0x1203, 0x1204, 0x1205, 0x1206,     -1,
     0x1208, 0x1209, 0x120a, 0x120b, 0x120c, 0x120d, 0x120e, 0x120f,
     0x1210, 0x1211, 0x1212, 0x1213, 0x1214, 0x1215, 0x1216, 0x1217,
     0x1218, 0x1219, 0x121a, 0x121b, 0x121c, 0x121d, 0x121e, 0x121f,
     0x1220, 0x1221, 0x1222, 0x1223, 0x1224, 0x1225, 0x1226, 0x1227,
     0x1228, 0x1229, 0x122a, 0x122b, 0x122c, 0x122d, 0x122e, 0x122f,
     0x1230, 0x1231, 0x1232, 0x1233, 0x1234, 0x1235, 0x1236, 0x1237,
     0x1238, 0x1239, 0x123a, 0x123b, 0x123c, 0x123d, 0x123e, 0x123f,
     0x1240, 0x1241, 0x1242, 0x1243, 0x1244, 0x1245, 0x1246,     -1,
     0x1248,     -1, 0x124a, 0x124b, 0x124c, 0x124d,     -1,     -1,
     0x1250, 0x1251, 0x1252, 0x1253, 0x1254, 0x1255, 0x1256,     -1,
     0x1258,     -1, 0x125a, 0x125b, 0x125c, 0x125d,     -1,     -1,
     0x1260, 0x1261, 0x1262, 0x1263, 0x1264, 0x1265, 0x1266, 0x1267,
     0x1268, 0x1269, 0x126a, 0x126b, 0x126c, 0x126d, 0x126e, 0x126f,
     0x1270, 0x1271, 0x1272, 0x1273, 0x1274, 0x1275, 0x1276, 0x1277,
     0x1278, 0x1279, 0x127a, 0x127b, 0x127c, 0x127d, 0x127e, 0x127f,
     0x1280, 0x1281, 0x1282, 0x1283, 0x1284, 0x1285, 0x1286,     -1,
     0x1288,     -1, 0x128a, 0x128b, 0x128c, 0x128d,     -1,     -1,
     0x1290, 0x1291, 0x1292, 0x1293, 0x1294, 0x1295, 0x1296, 0x1297,
     0x1298, 0x1299, 0x129a, 0x129b, 0x129c, 0x129d, 0x129e, 0x129f,
     EMPTYLINE, EMPTYLINE, EMPTYLINE, EMPTYLINE, 
};

static int inuit[224] = { /* no need to check for them all... */
         -1,     -1,     -1, 0x1403, 0x1404, 0x1405, 0x1406,     -1,
     0x1431, 0x1432, 0x1433, 0x1434, 0x1438, 0x1439, 0x144e, 0x144f,
     0x1450, 0x1451, 0x1455, 0x1456, 0x146d, 0x146e, 0x146f, 0x1470,
     0x1472, 0x1473, 0x148b, 0x148c, 0x148d, 0x148e, 0x1490, 0x1491,
     0x14a5, 0x14a6, 0x14a7, 0x14a8, 0x14aa, 0x14ab, 0x14c2, 0x14c3,
     0x14c4, 0x14c5, 0x14c7, 0x14c8, 0x14d5, 0x14d6, 0x14d7, 0x14d8,
     0x14da, 0x14db, 0x14ef, 0x14f0, 0x14f1, 0x14f2, 0x14f4, 0x14f5,
     0x1528, 0x1529, 0x152a, 0x152b, 0x152d, 0x152e, 0x1546, 0x1547,
	 EMPTYLINE, EMPTYLINE, EMPTYLINE, EMPTYLINE, EMPTYLINE,
     EMPTYLINE, EMPTYLINE, EMPTYLINE, EMPTYLINE, EMPTYLINE,
};


static int A_Z_letters[224] = {
    0x0020,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047, 
    0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f, 
    0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057, 
    0x0058, 0x0059, 0x005a,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1, 
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
};

static int viscii[224] = {
/* 0x0020 - 0x007f */
    0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027,
    0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f,
    0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037,
    0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f,
    0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047,
    0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f,
    0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057,
    0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x005e, 0x005f,
    0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067,
    0x0068, 0x0069, 0x006a, 0x006b, 0x006c, 0x006d, 0x006e, 0x006f,
    0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077,
    0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d, 0x007e, ALTCHR,
/* 0x0080 - 0x009F */
    0x1ea0, 0x1eae, 0x1eb0, 0x1eb6, 0x1ea4, 0x1ea6, 0x1ea8, 0x1eac,
    0x1ebc, 0x1eb8, 0x1ebe, 0x1ec0, 0x1ec2, 0x1ec4, 0x1ec6, 0x1ed0,
    0x1ed2, 0x1ed4, 0x1ed6, 0x1ed8, 0x1ee2, 0x1eda, 0x1edc, 0x1ede,
    0x1ec4, 0x1ece, 0x1ecc, 0x1ec8, 0x1ec6, 0x0168, 0x1ee4, 0x1ef2,
/* 0x00A0 - 0x00FF */
    0x00d5, 0x1eaf, 0x1eb1, 0x1eb7, 0x1ea5, 0x1ea7, 0x1ea9, 0x1ead,
    0x1ebd, 0x1eb9, 0x1ebf, 0x1ec1, 0x1ec3, 0x1ec5, 0x1ec7, 0x1ed1,
    0x1ed3, 0x1ed5, 0x1ed7, 0x1ee0, 0x01a0, 0x1ed9, 0x1edc, 0x1edf,
    0x1ec5, 0x1ef0, 0x1ee8, 0x1eea, 0x1eec, 0x01a1, 0x1edb, 0x01af,
    0x00c0, 0x00c1, 0x00c2, 0x00c3, 0x1ea2, 0x0102, 0x1eb3, 0x1eb5,
    0x00c8, 0x00c9, 0x00ca, 0x1eb4, 0x00cc, 0x00cd, 0x0128, 0x1ef3,
    0x0110, 0x1ee9, 0x00d2, 0x00d3, 0x00d4, 0x1ea1, 0x1ef7, 0x1eeb,
    0x1eed, 0x00d9, 0x00da, 0x1ef9, 0x1ef5, 0x00dd, 0x1ee1, 0x01b0,
    0x00e0, 0x00e1, 0x00e2, 0x00e3, 0x1ea3, 0x0103, 0x1eef, 0x1eab,
    0x00e8, 0x00e9, 0x00ea, 0x1eb5, 0x00ec, 0x00ed, 0x0129, 0x1ec9,
    0x0111, 0x1ef1, 0x00f2, 0x00f3, 0x00f4, 0x00f5, 0x1ecf, 0x1ecd,
    0x1ee5, 0x00f9, 0x00fa, 0x0169, 0x1ee7, 0x00fd, 0x1ee3, 0x1eee
};

static int tcvn[224] = {
/* 0x0020 - 0x007f */
    0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027,
    0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f,
    0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037,
    0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f,
    0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047,
    0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f,
    0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057,
    0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x005e, 0x005f,
    0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067,
    0x0068, 0x0069, 0x006a, 0x006b, 0x006c, 0x006d, 0x006e, 0x006f,
    0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077,
    0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d, 0x007e, ALTCHR,
/* 0x0080 - 0x009F */
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
/* 0x00a0 - 0x00ff */
    ALTCHR, 0x0102, 0x00c2, 0x00ca, 0x00d4, 0x01a0, 0x01af, 0x0110,
    0x0103, 0x00e2, 0x00ea, 0x00f4, 0x01a1, 0x01b0, 0x0111, ALTCHR,
    ALTCHR, ALTCHR, ALTCHR, ALTCHR, ALTCHR, 0x00e0, 0x1ea3, 0x00e3,
    0x00e1, 0x1ea1, ALTCHR, 0x1eb1, 0x1eb3, 0x1eb5, 0x1eaf, ALTCHR,
    ALTCHR, ALTCHR, ALTCHR, ALTCHR, ALTCHR, ALTCHR, 0x1eb7, 0x1ea7,
    0x1ea9, 0x1eab, 0x1ea5, 0x1ead, 0x00e8, ALTCHR, 0x1eb5, 0x1ebd,
    0x00e9, 0x1eb9, 0x1ec1, 0x1ec3, 0x1ec5, 0x1eb7, 0x1ec7, 0x00ec,
    0x1ec9, ALTCHR, ALTCHR, ALTCHR, 0x0129, 0x00ed, 0x1ec5, 0x00f2,
    ALTCHR, 0x1ecf, 0x00f5, 0x00f3, 0x1ecd, 0x1ed3, 0x1ed5, 0x1ed7,
    0x1ed1, 0x1ed9, 0x1edc, 0x1edf, 0x1ee1, 0x1edb, 0x1ee3, 0x00f9,
    ALTCHR, 0x1ee7, 0x0169, 0x00fa, 0x1ee5, 0x1eeb, 0x1eed, 0x1eef,
    0x1ee9, 0x1ef1, 0x1ef3, 0x1ef7, 0x1ef9, 0x00fd, 0x1ef5, ALTCHR
};

static int armscii[224] = {
/* 0x0020 - 0x007f */
    0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027,
    0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f,
    0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037,
    0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f,
    0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047,
    0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f,
    0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057,
    0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x005e, 0x005f,
    0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067,
    0x0068, 0x0069, 0x006a, 0x006b, 0x006c, 0x006d, 0x006e, 0x006f,
    0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077,
    0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d, 0x007e, ALTCHR,
/* 0x0080 - 0x009F */
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
/* 0x00a0 - 0x00ff */
    0x00a0, 0xf0a1, 0x00a7, 0x0589, 0x0029, 0x0028, 0x00bb, 0x00ab,
    0x2014, 0x002e, 0x055d, 0x002c, 0x2010, 0x058a, 0x2026, 0x055c,
    0x055b, 0x055e, 0x0531, 0x0561, 0x0532, 0x0562, 0x0533, 0x0563,
    0x0534, 0x0564, 0x0535, 0x0565, 0x0536, 0x0566, 0x0537, 0x0567,
    0x0538, 0x0568, 0x0539, 0x0569, 0x053a, 0x056a, 0x053b, 0x056b,
    0x053c, 0x056c, 0x053d, 0x056d, 0x053e, 0x056e, 0x053f, 0x056f,
    0x0540, 0x0570, 0x0541, 0x0571, 0x0542, 0x0572, 0x0543, 0x0573,
    0x0544, 0x0574, 0x0545, 0x0575, 0x0546, 0x0576, 0x0547, 0x0577,
    0x0548, 0x0578, 0x0549, 0x0579, 0x054a, 0x057a, 0x054b, 0x057b,
    0x054c, 0x057c, 0x054d, 0x057d, 0x054e, 0x057e, 0x054f, 0x057f,
    0x0550, 0x0580, 0x0551, 0x0581, 0x0552, 0x0582, 0x0553, 0x0583,
    0x0554, 0x0584, 0x0555, 0x0585, 0x0556, 0x0586, 0x02bc, 0x0587
};

static int georgianrs[224] = {
/* 0x0020 - 0x007f */
    0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027,
    0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f,
    0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037,
    0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f,
    0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047,
    0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f,
    0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057,
    0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x005e, 0x005f,
    0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067,
    0x0068, 0x0069, 0x006a, 0x006b, 0x006c, 0x006d, 0x006e, 0x006f,
    0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077,
    0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d, 0x007e, ALTCHR,
/* 0x0080 - 0x009F */
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
/* 0x00a0 - 0x00ff */
    0x00a0, 0x00a1, 0x00a2, 0x00a3, 0x00a4, 0x00a5, 0x00a6, 0x00a7,
    0x00a8, 0x00a9, 0x00aa, 0x00ab, 0x00ac, 0x00ad, 0x00ae, 0x00af,
    0x00b0, 0x00b1, 0x00b2, 0x00b3, 0x00b4, 0x00b5, 0x00b6, 0x00b7,
    0x00b8, 0x00b9, 0x00ba, 0x00bb, 0x00bc, 0x00bd, 0x00be, 0x00bf,
    0x10d0, 0x10d1, 0x10d2, 0x10d3, 0x10d4, 0x10d5, 0x10d6, 0x10d7,
    0x10d8, 0x10d9, 0x10da, 0x10db, 0x10dc, 0x10dd, 0x10de, 0x10df,
    0x10e0, 0x10e1, 0x10e2, 0x10e3, 0x10e4, 0x10e5, 0x10e6, 0x10e7,
    0x10e8, 0x10e9, 0x10ea, 0x10eb, 0x10ec, 0x10ed, 0x10ee, 0x10ef,
    0x10f0, 0x10f1, 0x10f2, 0x10f3, 0x10f4, 0x10f5, 0x10f6,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1
};

static int georgianstd[224] = {
/* 0x0020 - 0x007f */
    0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027,
    0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f,
    0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037,
    0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f,
    0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047,
    0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f,
    0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057,
    0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x005e, 0x005f,
    0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067,
    0x0068, 0x0069, 0x006a, 0x006b, 0x006c, 0x006d, 0x006e, 0x006f,
    0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077,
    0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d, 0x007e, ALTCHR,
/* 0x0080 - 0x009F */
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
/* 0x00a0 - 0x00ff */
    0x00a0, 0x00a1, 0x00a2, 0x00a3, 0x00a4, 0x00a5, 0x00a6, 0x00a7,
    0x00a8, 0x00a9, 0x00aa, 0x00ab, 0x00ac, 0x00ad, 0x00ae, 0x00af,
    0x00b0, 0x00b1, 0x00b2, 0x00b3, 0x00b4, 0x00b5, 0x00b6, 0x00b7,
    0x00b8, 0x00b9, 0x00ba, 0x00bb, 0x00bc, 0x00bd, 0x00be, 0x00bf,
    0x10d0, 0x10d1, 0x10d2, 0x10d3, 0x10d4, 0x10d5, 0x10d6, 0x10f5,
    0x10d7, 0x10d8, 0x10d9, 0x10da, 0x10db, 0x10dc, 0x10f2, 0x10dd,
    0x10de, 0x10df, 0x10e0, 0x10e1, 0x10e2, 0x10f3, 0x10e3, 0x10e4,
    0x10e5, 0x10e6, 0x10e7, 0x10e8, 0x10e9, 0x10ea, 0x10eb, 0x10ec,
    0x10ed, 0x10ee, 0x10f1, 0x10ef, 0x10f0, 0x10f4, 0x10f6,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1
};


static int mulelao[224] = {
/* 0x0020 - 0x007f */
    0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027,
    0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f,
    0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037,
    0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f,
    0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047,
    0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f,
    0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057,
    0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x005e, 0x005f,
    0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067,
    0x0068, 0x0069, 0x006a, 0x006b, 0x006c, 0x006d, 0x006e, 0x006f,
    0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077,
    0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d, 0x007e, ALTCHR,
/* 0x0080 - 0x009F */
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
/* 0x00A0 - 0x00FF */
    ALTCHR, 0x0E81, 0x0E82, ALTCHR, 0x0E84, ALTCHR, ALTCHR, 0x0E87,
    0x0E88, ALTCHR, 0x0E8A, ALTCHR, ALTCHR, 0x0E8D, ALTCHR, ALTCHR,
    ALTCHR, ALTCHR, ALTCHR, ALTCHR, 0x0E94, 0x0E95, 0x0E96, 0x0E97,
    ALTCHR, 0x0E99, 0x0E9A, 0x0E9B, 0x0E9C, 0x0E9D, 0x0E9E, 0x0E9F,
    ALTCHR, 0x0EA1, 0x0EA2, 0x0EA3, ALTCHR, 0x0EA5, ALTCHR, 0x0EA7,
    ALTCHR, ALTCHR, 0x0EAA, 0x0EAB, ALTCHR, 0x0EAD, 0x0EAE, 0x0EAF,
    0x0EB0, 0x0EB1, 0x0EB2, 0x0EB3, 0x0EB4, 0x0EB5, 0x0EB6, 0x0EB7,
    0x0EB8, 0x0EB9, ALTCHR, 0x0EBB, 0x0EBC, 0x0EBD, ALTCHR, 0x20ad,
    0x0EC0, 0x0EC1, 0x0EC2, 0x0EC3, 0x0EC4, ALTCHR, 0x0EC6, ALTCHR,
    0x0EC8, 0x0EC9, 0x0ECA, 0x0ECB, 0x0ECC, 0x0ECD, ALTCHR, ALTCHR,
    0x0ED0, 0x0ED1, 0x0ED2, 0x0ED3, 0x0ED4, 0x0ED5, 0x0ED6, 0x0ED7,
    0x0ED8, 0x0ED9, ALTCHR, 0x0EDC, 0x0EDD, ALTCHR, ALTCHR, ALTCHR
};

static int asciilao[224] = {
/* 0x0020 - 0x007f */
    0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027,
    0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f,
    0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037,
    0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f,
    0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047,
    0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f,
    0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057,
    0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x005e, 0x005f,
    0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067,
    0x0068, 0x0069, 0x006a, 0x006b, 0x006c, 0x006d, 0x006e, 0x006f,
    0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077,
    0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d, 0x007e, ALTCHR,
/* 0x0080 - 0x009F */
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
/* 0x00a0 - 0x00ff */
    ALTCHR, 0x0E81, 0x0E82, 0x0E84, 0x0E87, 0x0E88, 0x0EAA, 0x0E8A,
    0x0E8D, 0x0E94, 0x0E95, 0x0E96, 0x0E97, 0x0E99, 0x0E9A, 0x0E9B,
    0x0E9C, 0x0E9D, 0x0E9E, 0x0E9F, 0x0EA1, 0x0EA2, 0x0EA3, 0x0EA5,
    0x0EA7, 0x0EAB, 0x0EAD, 0x0EAE, ALTCHR, ALTCHR, ALTCHR, 0x0EAF,
    0x0EB0, 0x0EB2, 0x0EB3, 0x0EB4, 0x0EB5, 0x0EB6, 0x0EB7, 0x0EB8,
    0x0EB9, 0x0EBC, 0x0EB1, 0x0EBB, 0x0EBD, ALTCHR, ALTCHR, ALTCHR,
    0x0EC0, 0x0EC1, 0x0EC2, 0x0EC3, 0x0EC4, 0x0EC8, 0x0EC9, 0x0ECA,
    0x0ECB, 0x0ECC, 0x0ECD, 0x0EC6, ALTCHR, 0x0EDC, 0x0EDD, 0x20AD,
    ALTCHR, ALTCHR, ALTCHR, ALTCHR, ALTCHR, ALTCHR, ALTCHR, ALTCHR,
    ALTCHR, ALTCHR, ALTCHR, ALTCHR, ALTCHR, ALTCHR, ALTCHR, ALTCHR,
    0x0ED0, 0x0ED1, 0x0ED2, 0x0ED3, 0x0ED4, 0x0ED5, 0x0ED6, 0x0ED7,
    0x0ED8, 0x0ED9, ALTCHR, ALTCHR, 0x00A2, 0x00AC, 0x00A6, ALTCHR
};

static int jisx0201[224] = {
    0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027,
    0x0028, 0x0029, 0x002A, 0x002B, 0x002C, 0x002D, 0x002E, 0x002F,
    0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037,
    0x0038, 0x0039, 0x003A, 0x003B, 0x003C, 0x003D, 0x003E, 0x003F,
    0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047,
    0x0048, 0x0049, 0x004A, 0x004B, 0x004C, 0x004D, 0x004E, 0x004F,
    0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057,
    0x0058, 0x0059, 0x005A, 0x005B, 0x00A5, 0x005D, 0x005E, 0x005F,
    0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067,
    0x0068, 0x0069, 0x006A, 0x006B, 0x006C, 0x006D, 0x006E, 0x006F,
    0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077,
    0x0078, 0x0079, 0x007A, 0x007B, 0x007C, 0x007D, 0x007E,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1, 0xFF61, 0xFF62, 0xFF63, 0xFF64, 0xFF65, 0xFF66, 0xFF67,
    0xFF68, 0xFF69, 0xFF6A, 0xFF6B, 0xFF6C, 0xFF6D, 0xFF6E, 0xFF6F,
    0xFF70, 0xFF71, 0xFF72, 0xFF73, 0xFF74, 0xFF75, 0xFF76, 0xFF77,
    0xFF78, 0xFF79, 0xFF7A, 0xFF7B, 0xFF7C, 0xFF7D, 0xFF7E, 0xFF7F,
    0xFF80, 0xFF81, 0xFF82, 0xFF83, 0xFF84, 0xFF85, 0xFF86, 0xFF87,
    0xFF88, 0xFF89, 0xFF8A, 0xFF8B, 0xFF8C, 0xFF8D, 0xFF8E, 0xFF8F,
    0xFF90, 0xFF91, 0xFF92, 0xFF93, 0xFF94, 0xFF95, 0xFF96, 0xFF97,
    0xFF98, 0xFF99, 0xFF9A, 0xFF9B, 0xFF9C, 0xFF9D, 0xFF9E, 0xFF9F,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1,
    /**/-1,     -1,     -1,     -1,     -1,     -1,     -1,     -1
};

static int jisx0208[224] = {
/* only a very limited subset of it; hope it is enough to check */
    0x70CF, 0x7FBD, 0x8FC2, 0x96E8, 0x536F, 0x9D5C, 0x7ABA, 0x4E11,
    0x7893, 0x81FC, 0x6E26, 0x5618, 0x5504, 0x6B1D, 0x851A, 0x9C3B,
    0x59E5, 0x53A9, 0x6D66, 0x74DC, 0x958F, 0x5642, 0x4E91, 0x904B,
    0x96F2, 0x834F, 0x990C, 0x53E1, 0x55B6, 0x5B30, 0x5F71, 0x6620,
    0x66F3, 0x6804, 0x6C38, 0x6CF3, 0x6D29, 0x745B, 0x76C8, 0x7A4E,
    0x9834, 0x82F1, 0x885B, 0x8A60, 0x92ED, 0x6DB2, 0x75AB, 0x76CA,
    0x99C5, 0x60A6, 0x8B01, 0x8D8A, 0x95B2, 0x698E, 0x53AD, 0x5186,
    0x5712, 0x5830, 0x5944, 0x5BB4, 0x5EF6, 0x6028, 0x63A9, 0x63F4,
    0x6CBF, 0x6F14, 0x708E, 0x7114, 0x7159, 0x71D5, 0x733F, 0x7E01,
    0x8276, 0x82D1, 0x8597, 0x9060, 0x925B, 0x9D1B, 0x5869, 0x65BC,
    0x7D75, 0x82A5, 0x87F9, 0x958B, 0x968E, 0x8C9D, 0x51F1, 0x52BE,
    0x5916, 0x54B3, 0x5BB3, 0x5D16, 0x6168, 0x6982, 0x6DAF, 0x788D,
    0x84CB, 0x8857, 0x8A72, 0x93A7, 0x9AB8, 0x6D6C, 0x99A8, 0x86D9,
    0x57A3, 0x67FF, 0x86CE, 0x920E, 0x5283, 0x5687, 0x5404, 0x5ED3,
    0x62E1, 0x64B9, 0x683C, 0x6838, 0x6BBB, 0x7372, 0x78BA, 0x7A6B,
    0x899A, 0x89D2, 0x8D6B, 0x8F03, 0x90ED, 0x95A3, 0x9694, 0x9769,
    0x5B66, 0x5CB3, 0x697D, 0x984D, 0x984E, 0x639B, 0x7B20, 0x6A2B,
    0x6A7F, 0x68B6, 0x9C0D, 0x6F5F, 0x5272, 0x559D, 0x6070, 0x62EC,
    0x6D3B, 0x6E07, 0x6ED1, 0x845B, 0x8910, 0x8F44, 0x4E14, 0x9C39,
    0x53F6, 0x691B, 0x6A3A, 0x9784, 0x682A, 0x515C, 0x7AC3, 0x84B2,
    0x5B63, 0x7A00, 0x7D00, 0x5FBD, 0x898F, 0x8A18, 0x8CB4, 0x8D77,
    0x8ECC, 0x8F1D, 0x98E2, 0x9A0E, 0x9B3C, 0x4E80, 0x507D, 0x5100,
    0x5993, 0x5B9C, 0x622F, 0x6280, 0x64EC, 0x6B3A, 0x72A0, 0x7591,
    0x7947, 0x7FA9, 0x87FB, 0x8ABC, 0x8B70, 0x63AC, 0x83CA, 0x97A0,
    0x5B8F, 0x5DE5, 0x5DE7, 0x5DF7, 0x5E78, 0x5E83, 0x5E9A, 0x5EB7,
    0x5F18, 0x6052, 0x614C, 0x6297, 0x62D8, 0x63A7, 0x653B, 0x6602,
    0x6643, 0x66F4, 0x676D, 0x6821, 0x6897, 0x69CB, 0x6C5F, 0x6D2A,
    0x6D69, 0x6E2F, 0x6E9D, 0x7532, 0x7687, 0x786C, 0x7A3F, 0x7CE0
};

static int jisx0212[224] = {
/* only a very limited subset of it; hope it is enough to check */
    0x0408, 0x0409, 0x040A, 0x040B, 0x040C, 0x040E, 0x040F, 0x0405,
    0x4E24, 0x4E28, 0x4E2B, 0x4E2E, 0x4E2F, 0x4E30, 0x4E35, 0x4E40,
    0x4E41, 0x4E44, 0x4E47, 0x4E51, 0x4E5A, 0x4E5C, 0x4E63, 0x4E68,
    0x4E69, 0x4E74, 0x4E75, 0x4E79, 0x4E7F, 0x4E8D, 0x4E96, 0x4E97,
    0x4E9D, 0x4EAF, 0x4EB9, 0x4EC3, 0x4ED0, 0x4EDA, 0x4EDB, 0x4EE0,
    0x4EE1, 0x4EE2, 0x4EE8, 0x4EEF, 0x4EF1, 0x4EF3, 0x4EF5, 0x4EFD,
    0x4EFE, 0x4EFF, 0x4F00, 0x4F02, 0x4F03, 0x4F08, 0x4F0B, 0x4F0C,
    0x4F12, 0x4F15, 0x4F16, 0x4F17, 0x4F19, 0x4F2E, 0x4F31, 0x4F60,
    0x4F33, 0x4F35, 0x4F37, 0x4F39, 0x4F3B, 0x4F3E, 0x4F40, 0x4F42,
    0x4F48, 0x4F49, 0x4F4B, 0x4F4C, 0x4F52, 0x4F54, 0x4F56, 0x4F58,
    0x509F, 0x50A0, 0x50A1, 0x50A2, 0x50AA, 0x50AF, 0x50B0, 0x50B9,
    0x50E4, 0x50E6, 0x50E8, 0x50E9, 0x50EF, 0x50F1, 0x50F6, 0x50FA,
    0x50FE, 0x5103, 0x5106, 0x5107, 0x5108, 0x510B, 0x510C, 0x510D,
    0x510E, 0x50F2, 0x5110, 0x5117, 0x5119, 0x511B, 0x511C, 0x511D,
    0x511E, 0x5123, 0x5127, 0x5128, 0x512C, 0x512D, 0x512F, 0x5131,
    0x5133, 0x5134, 0x5135, 0x5138, 0x5139, 0x5142, 0x514A, 0x514F,
    0x5153, 0x5155, 0x5157, 0x5158, 0x515F, 0x5164, 0x5166, 0x517E,
    0x5878, 0x587C, 0x587F, 0x5880, 0x5881, 0x5887, 0x5888, 0x5889,
    0x588A, 0x588C, 0x588D, 0x588F, 0x5890, 0x5894, 0x5896, 0x589D,
    0x58A0, 0x58A1, 0x58A2, 0x58A6, 0x58A9, 0x58B1, 0x58B2, 0x58C4,
    0x58BC, 0x58C2, 0x58C8, 0x58CD, 0x58CE, 0x58D0, 0x58D2, 0x58D4,
    0x58D6, 0x58DA, 0x58DD, 0x58E1, 0x58E2, 0x58E9, 0x58F3, 0x5905,
    0x6213, 0x6215, 0x621C, 0x6220, 0x6222, 0x6223, 0x6227, 0x6229,
    0x622B, 0x6239, 0x623D, 0x6242, 0x6243, 0x6244, 0x6246, 0x624C,
    0x71D6, 0x71D9, 0x71DA, 0x71DC, 0x71F8, 0x71FE, 0x7200, 0x7207,
    0x7208, 0x7209, 0x7213, 0x7217, 0x721A, 0x721D, 0x721F, 0x7224,
    0x802C, 0x802E, 0x8030, 0x8034, 0x8035, 0x8037, 0x8039, 0x803A,
    0x803C, 0x803E, 0x8040, 0x8044, 0x8060, 0x8064, 0x8066, 0x806D
};

static int ksc5601[224] = {
/* only a very limited subset of it; hope it is enough to check */
    0x2033, 0x2103, 0x212B, 0xFFE0, 0xFFE1, 0xFFE5, 0x2642, 0x2640,
    0x2220, 0x22A5, 0x2312, 0x2202, 0x2207, 0x2261, 0x2252, 0x00A7,
    0x3150, 0x3151, 0x3152, 0x3153, 0x3154, 0x3155, 0x3156, 0x3157,
    0x3158, 0x3159, 0x315A, 0x315B, 0x315C, 0x315D, 0x315E, 0x315F,
    0x3160, 0x3161, 0x3162, 0x3163, 0x3164, 0x3165, 0x3166, 0x3167,
    0x3168, 0x3169, 0x316A, 0x316B, 0x316C, 0x316D, 0x316E, 0x316F,
    0x3050, 0x3051, 0x3052, 0x3053, 0x3054, 0x3055, 0x3056, 0x3057,
    0x3058, 0x3059, 0x305A, 0x305B, 0x305C, 0x305D, 0x305E, 0x305F,
    0xAD6C, 0xAD6D, 0xAD70, 0xAD73, 0xAD74, 0xAD75, 0xAD76, 0xAD7B,
    0xAD7C, 0xAD7D, 0xAD7F, 0xAD81, 0xAD82, 0xAD88, 0xAD89, 0xAD8C,
    0xAD90, 0xAD9C, 0xAD9D, 0xADA4, 0xADB7, 0xADC0, 0xADC1, 0xADC4,
    0xADC8, 0xADD0, 0xADD1, 0xADD3, 0xADDC, 0xADE0, 0xADE4, 0xADF8,
    0xB394, 0xB3A0, 0xB3A1, 0xB3A8, 0xB3AC, 0xB3C4, 0xB3C5, 0xB3C8,
    0xB3CB, 0xB3CC, 0xB3CE, 0xB3D0, 0xB3D4, 0xB3D5, 0xB3D7, 0xB3D9,
    0xB3DB, 0xB3DD, 0xB3E0, 0xB3E4, 0xB3E8, 0xB3FC, 0xB410, 0xB418,
    0xB41C, 0xB420, 0xB428, 0xB429, 0xB42B, 0xB434, 0xB450, 0xB451,
    0xB807, 0xB808, 0xB809, 0xB80C, 0xB810, 0xB818, 0xB819, 0xB81B,
    0xB81D, 0xB824, 0xB825, 0xB828, 0xB82C, 0xB834, 0xB835, 0xB837,
    0xBE7B, 0xBE7C, 0xBE7D, 0xBE80, 0xBE84, 0xBE8C, 0xBE8D, 0xBE8F,
    0xBE90, 0xBE91, 0xBE98, 0xBE99, 0xBEA8, 0xBED0, 0xBED1, 0xBED4,
    0xBED7, 0xBED8, 0xBEE0, 0xBEE3, 0xBEE4, 0xBEE5, 0xBEEC, 0xBF01,
    0xBF08, 0xBF09, 0xBF18, 0xBF19, 0xBF1B, 0xBF1C, 0xBF1D, 0xBF40,
    0xC27C, 0xC27D, 0xC27F, 0xC281, 0xC288, 0xC289, 0xC290, 0xC298,
    0xC29B, 0xC29D, 0xC2A4, 0xC2A5, 0xC2A8, 0xC2AC, 0xC2AD, 0xC2B4,
    0x54E5, 0x5609, 0x5AC1, 0x5BB6, 0x6687, 0x67B6, 0x67B7, 0x67EF,
    0x6B4C, 0x73C2, 0x75C2, 0x7A3C, 0x82DB, 0x8304, 0x8857, 0x8888,
    0x7FA4, 0x88D9, 0x8ECD, 0x90E1, 0x5800, 0x5C48, 0x6398, 0x7A9F,
    0x5BAE, 0x5F13, 0x7A79, 0x7AAE, 0x828E, 0x8EAC, 0x5026, 0x5238
};

static int johab[224] = {
/* only a very limited subset of it; hope it is enough to check */
    0xAC02, 0xAC03, 0xAC25, 0xAC2B, 0xAC53, 0xAC57, 0xAD98, 0xAD9B,
    0xADA8, 0xADAF, 0xADC7, 0xADDE, 0xB9AF, 0xB9B2, 0xB9C2, 0xB9CB,
    0xBC04, 0xBC05, 0xBD00, 0xBD01, 0xBDE2, 0xBDE3, 0xBEC7, 0xBEC8, 
    0xBFA9, 0xBFB4, 0xC08F, 0xC092, 0xC14E, 0xC14F, 0xC28B, 0xC28C, 
    0xC352, 0xC353, 0xC46B, 0xC4AB, 0xC4C1, 0xC5EF, 0xC627, 0xC6B6, 
    0xC6D1, 0xC7C5, 0xC7D7, 0xC8AC, 0xC8B9, 0xC912, 0xC976, 0xC977, 
    0xC9E6, 0xC9E8, 0xCA51, 0xCA52, 0xCAAB, 0xCAAC, 0xCABA, 0xCABB, 
    0xCABE, 0xCABF, 0xCAC1, 0xCAC2, 0xCAC3, 0xCAC5, 0xCAC6, 0xCB16, 
    0xCB17, 0xCB19, 0xCB70, 0xCB7D, 0xCBCE, 0xCBD6, 0xCBD7, 0xCBD8, 
    0xCC26, 0xCC95, 0xCC9E, 0xCCFF, 0xCD62, 0xCD63, 0xCDC7, 0xCE34, 
    0xCE9B, 0xCF03, 0xCF6E, 0xCFCD, 0xD032, 0xD033, 0xD036, 0xD098, 
    0xD111, 0xD112, 0xD171, 0xD1D1, 0xD22B, 0xD28E, 0xD2FC, 0xD2FD, 
    0xD304, 0xD36D, 0xD3D9, 0xD3DA, 0xD439, 0xD49F, 0xD4A0, 0xD501, 
    0xD502, 0xD566, 0xD567, 0xD56A, 0xD56C, 0xD56E, 0xD56F, 0xD570, 
    0xD571, 0xD572, 0xD573, 0xD576, 0xD577, 0xD579, 0xD57A, 0xD57B, 
    0xD57D, 0xD57E, 0xD57F, 0xD5CB, 0xD63E, 0xD63F, 0xD641, 0xD642, 
    0xD643, 0xD644, 0xD646, 0xD647, 0xD64A, 0xD64C, 0xD64E, 0xD64F, 
    0xD650, 0xD652, 0xD653, 0xD656, 0xD657, 0xD659, 0xD65A, 0xD65B, 
    0xD65D, 0xD65E, 0xD65F, 0xD660, 0xD661, 0xD662, 0xD663, 0xD664, 
    0xD665, 0xD666, 0xD668, 0xD66A, 0xD66B, 0xD66C, 0xD66D, 0xD675, 
    0xD676, 0xD677, 0xD678, 0xD679, 0xD67C, 0xD67D, 0xD67E, 0xD67F, 
    0xD680, 0xD681, 0xD682, 0xD684, 0xD686, 0xD687, 0xD688, 0xD689,  
    0xD68B, 0xD68E, 0xD68F, 0xD691, 0xD692, 0xD693, 0xD695, 0xD696, 
    0xD697, 0xD698, 0xD699, 0xD69A, 0xD69B, 0xD69C, 0xD69E, 0xD6A0, 
    0xD6A2, 0xD6A3, 0xD6A4, 0xD6A5, 0xD6A6, 0xD6A7, 0xD6A9, 0xD6AA, 
    0xD6AB, 0xD6AD, 0xD6AE, 0xD6AF, 0xD6B1, 0xD6B2, 0xD6B3, 0xD6B4, 
    0xD6B5, 0xD6B6, 0xD6B7, 0xD6B8, 0xD6BA, 0xD6BC, 0xD6BD, 0xD6BE, 
    0xD6D2, 0xD6D3, 0xD6D5, 0xD6D6, 0xD6D8, 0xD6DA, 0xD6DB, 0xD6DC 
};

static int gb2312[224] = {
/* only a very limited subset of it; hope it is enough to check */
    0x2312, 0x2299, 0x222B, 0x222E, 0x2261, 0x224C, 0x2248, 0x223D,
    0x221D, 0x2260, 0x226E, 0x226F, 0x2264, 0x2265, 0x221E, 0x2235,
    0x2497, 0x2498, 0x2499, 0x249A, 0x249B, 0x2474, 0x2475, 0x2476,
    0x2477, 0x2478, 0x2479, 0x247A, 0x247B, 0x247C, 0x247D, 0x247E,
    0x3058, 0x3059, 0x305A, 0x305B, 0x305C, 0x305D, 0x305E, 0x305F,
    0x3060, 0x3061, 0x3062, 0x3063, 0x3064, 0x3065, 0x3066, 0x3067,
    0x03B8, 0x03B9, 0x03BA, 0x03BB, 0x03BC, 0x03BD, 0x03BE, 0x03BF,
    0x03C0, 0x03C1, 0x03C3, 0x03C4, 0x03C5, 0x03C6, 0x03C7, 0x03C8,
    0x6BD4, 0x9119, 0x7B14, 0x5F7C, 0x78A7, 0x84D6, 0x853D, 0x6BD5,
    0x6BD9, 0x6BD6, 0x5E01, 0x5E87, 0x75F9, 0x95ED, 0x655D, 0x5F0A,
    0x6A59, 0x6210, 0x5448, 0x4E58, 0x7A0B, 0x60E9, 0x6F84, 0x8BDA,
    0x627F, 0x901E, 0x9A8B, 0x79E4, 0x5403, 0x75F4, 0x6301, 0x5319,
    0x9492, 0x7E41, 0x51E1, 0x70E6, 0x53CD, 0x8FD4, 0x8303, 0x8D29,
    0x72AF, 0x996D, 0x6CDB, 0x574A, 0x82B3, 0x65B9, 0x80AA, 0x623F,
    0x94BE, 0x5047, 0x7A3C, 0x4EF7, 0x67B6, 0x9A7E, 0x5AC1, 0x6B7C,
    0x76D1, 0x575A, 0x5C16, 0x7B3A, 0x95F4, 0x714E, 0x517C, 0x80A9,
    0x5ECA, 0x90CE, 0x6717, 0x6D6A, 0x635E, 0x52B3, 0x7262, 0x8001,
    0x4F6C, 0x59E5, 0x916A, 0x70D9, 0x6D9D, 0x52D2, 0x4E50, 0x96F7,
    0x5C41, 0x8B6C, 0x7BC7, 0x504F, 0x7247, 0x9A97, 0x98D8, 0x6F02,
    0x74E2, 0x7968, 0x6487, 0x77A5, 0x62FC, 0x9891, 0x8D2B, 0x54C1,
    0x817F, 0x8715, 0x892A, 0x9000, 0x541E, 0x5C6F, 0x81C0, 0x62D6,
    0x6258, 0x8131, 0x9E35, 0x9640, 0x9A6E, 0x9A7C, 0x692D, 0x59A5,
    0x6FA1, 0x86A4, 0x8E81, 0x566A, 0x9020, 0x7682, 0x7076, 0x71E5,
    0x8D23, 0x62E9, 0x5219, 0x6CFD, 0x8D3C, 0x600E, 0x589E, 0x618E,
    0x539D, 0x53A3, 0x53A5, 0x53AE, 0x9765, 0x8D5D, 0x531A, 0x53F5,
    0x5326, 0x532E, 0x533E, 0x8D5C, 0x5366, 0x5363, 0x5202, 0x5208,
    0x9E92, 0x93D6, 0x9E9D, 0x9E9F, 0x9EDB, 0x9EDC, 0x9EDD, 0x9EE0,
    0x9EDF, 0x9EE2, 0x9EE9, 0x9EE7, 0x9EE5, 0x9EEA, 0x9EEF, 0x9F22
};

static int gbk[224] = {
/* only a very limited subset of it; hope it is enough to check */
    0x2312, 0x2299, 0x222B, 0x222E, 0x2261, 0x224C, 0x2248, 0x223D,
    0x221D, 0x2260, 0x2609, 0x226F, 0x2264, 0x2265, 0x221E, 0x2235,
    0x2497, 0x2498, 0x2499, 0x249A, 0x249B, 0x2474, 0x2475, 0x2476,
    0x2477, 0x2478, 0x2479, 0x247A, 0x247B, 0x247C, 0x247D, 0x247E,
    0x3058, 0x3059, 0x305A, 0x305B, 0x305C, 0x305D, 0x305E, 0x305F,
    0x3060, 0x3061, 0x3062, 0x3063, 0x3064, 0x3065, 0x3066, 0x3067,
    0x03B8, 0x03B9, 0x03BA, 0x03BB, 0x03BC, 0x03BD, 0x03BE, 0x03BF,
    0x03C0, 0x03C1, 0x03C3, 0x03C4, 0x03C5, 0x03C6, 0x03C7, 0x03C8,
    0x6BD4, 0x9119, 0x7B14, 0x5F7C, 0x78A7, 0x84D6, 0x853D, 0x6BD5,
    0x6BD9, 0x6BD6, 0x5E01, 0x5E87, 0x75F9, 0x95ED, 0x655D, 0x5F0A,
    0x6A59, 0x6210, 0x5448, 0x4E58, 0x7A0B, 0x60E9, 0x6F84, 0x8BDA,
    0x627F, 0x901E, 0x9A8B, 0x79E4, 0x5403, 0x75F4, 0x6301, 0x5319,
    0x9492, 0x7E41, 0x51E1, 0x70E6, 0x53CD, 0x8FD4, 0x8303, 0x8D29,
    0x72AF, 0x996D, 0x6CDB, 0x574A, 0x82B3, 0x65B9, 0x80AA, 0x623F,
    0x94BE, 0x5047, 0x7A3C, 0x4EF7, 0x67B6, 0x9A7E, 0x5AC1, 0x6B7C,
    0x76D1, 0x575A, 0x5C16, 0x7B3A, 0x95F4, 0x714E, 0x517C, 0x80A9,
    0x5ECA, 0x90CE, 0x6717, 0x6D6A, 0x635E, 0x52B3, 0x7262, 0x8001,
    0x4F6C, 0x59E5, 0x916A, 0x70D9, 0x6D9D, 0x52D2, 0x4E50, 0x96F7,
    0x5C41, 0x8B6C, 0x7BC7, 0x504F, 0x7247, 0x9A97, 0x98D8, 0x6F02,
    0x74E2, 0x7968, 0x6487, 0x77A5, 0x62FC, 0x9891, 0x8D2B, 0x54C1,
    0x817F, 0x8715, 0x892A, 0x9000, 0x541E, 0x5C6F, 0x81C0, 0x62D6,
    0x6258, 0x8131, 0x9E35, 0x9640, 0x9A6E, 0x9A7C, 0x692D, 0x59A5,
    0x6FA1, 0x86A4, 0x8E81, 0x566A, 0x9020, 0x7682, 0x7076, 0x71E5,
    0x8D23, 0x62E9, 0x5219, 0x6CFD, 0x8D3C, 0x600E, 0x589E, 0x618E,
    0x539D, 0x53A3, 0x53A5, 0x53AE, 0x9765, 0x8D5D, 0x531A, 0x53F5,
    0x5326, 0x532E, 0x533E, 0x8D5C, 0x5366, 0x5363, 0x5202, 0x5208,
    0x9E92, 0x93D6, 0x9E9D, 0x9E9F, 0x9EDB, 0x9EDC, 0x9EDD, 0x9EE0,
    0x3007, 0x3012, 0x3231, 0x32A3, 0x33A1, 0x33C4, 0x9EEF, 0x9F22
};

static int big5[224] = {
/* only a very limited subset of it; hope it is enough to check */
    0xFE36, 0xFF5B, 0xFF5D, 0xFE37, 0xFE38, 0x3014, 0x3015, 0xFE39,
    0xFE3A, 0x3010, 0x3011, 0xFE3B, 0xFE3C, 0x300A, 0x300B, 0xFE3D,
    0x5143, 0x5141, 0x5167, 0x516D, 0x516E, 0x516C, 0x5197, 0x51F6,
    0x5206, 0x5207, 0x5208, 0x52FB, 0x52FE, 0x52FF, 0x5316, 0x5339,
    0x52A3, 0x5308, 0x5321, 0x5320, 0x5370, 0x5371, 0x5409, 0x540F,
    0x540C, 0x540A, 0x5410, 0x5401, 0x540B, 0x5404, 0x5411, 0x540D,
    0x6C88, 0x6C89, 0x6C85, 0x6C9B, 0x6C6A, 0x6C7A, 0x6C90, 0x6C70,
    0x6C8C, 0x6C68, 0x6C96, 0x6C92, 0x6C7D, 0x6C83, 0x6C72, 0x6C7E,
    0x8010, 0x800D, 0x8011, 0x8036, 0x80D6, 0x80E5, 0x80DA, 0x80C3,
    0x80C4, 0x80CC, 0x80E1, 0x80DB, 0x80CE, 0x80DE, 0x80E4, 0x80DD,
    0x8A10, 0x8A0E, 0x8A0C, 0x8A15, 0x8A0A, 0x8A17, 0x8A13, 0x8A16,
    0x8A0F, 0x8A11, 0x8C48, 0x8C7A, 0x8C79, 0x8CA1, 0x8CA2, 0x8D77,
    0x5BD2, 0x5BCC, 0x5BD3, 0x5BD0, 0x5C0A, 0x5C0B, 0x5C31, 0x5D4C,
    0x5D50, 0x5D34, 0x5D47, 0x5DFD, 0x5E45, 0x5E3D, 0x5E40, 0x5E43,
    0x5AB3, 0x5AC2, 0x5AB2, 0x5D69, 0x5D6F, 0x5E4C, 0x5E79, 0x5EC9,
    0x5EC8, 0x5F12, 0x5F59, 0x5FAC, 0x5FAE, 0x611A, 0x610F, 0x6148,
    0x8A87, 0x8A7C, 0x8A63, 0x8AA0, 0x8A71, 0x8A85, 0x8A6D, 0x8A62,
    0x8A6E, 0x8A6C, 0x8A79, 0x8A7B, 0x8A3E, 0x8A68, 0x8C62, 0x8C8A,
    0x927C, 0x9291, 0x95A1, 0x95A8, 0x95A9, 0x95A3, 0x95A5, 0x95A4,
    0x9699, 0x969C, 0x969B, 0x96CC, 0x96D2, 0x9700, 0x977C, 0x9785,
    0x8ACB, 0x8AF8, 0x8AB2, 0x8AC9, 0x8AC2, 0x8ABF, 0x8AB0, 0x8AD6,
    0x8ACD, 0x8AB6, 0x8AB9, 0x8ADB, 0x8C4C, 0x8C4E, 0x8C6C, 0x8CE0,
    0x7E61, 0x7E52, 0x7E59, 0x7F48, 0x7FF9, 0x7FFB, 0x8077, 0x8076,
    0x81CD, 0x81CF, 0x820A, 0x85CF, 0x85A9, 0x85CD, 0x85D0, 0x85C9,
    0x880D, 0x87F9, 0x87FE, 0x8960, 0x895F, 0x8956, 0x895E, 0x8B41,
    0x8B5C, 0x8B58, 0x8B49, 0x8B5A, 0x8B4E, 0x8B4F, 0x8B46, 0x8B59,
    0x883F, 0x8B9E, 0x8C9C, 0x8EA9, 0x8EC9, 0x974B, 0x9873, 0x9874,
    0x98CC, 0x9961, 0x99AB, 0x9A64, 0x9A66, 0x9A67, 0x9B24, 0x9E15
};

static int mulearabic0[224] = {
/* 0x0020 - 0x002A */
    ALTCHR, 0x0660, 0x0661, 0x0662, 0x0663, 0x0664, 0x0665, 0x0666,
    0x0667, 0x0668, 0x0669, ALTCHR, ALTCHR, ALTCHR, ALTCHR, ALTCHR,
/* 0x0030 - 0x00FF */
    EMPTYLINE, EMPTYLINE, EMPTYLINE, EMPTYLINE, EMPTYLINE, EMPTYLINE,
    EMPTYLINE, EMPTYLINE, EMPTYLINE, EMPTYLINE, EMPTYLINE, EMPTYLINE,
    EMPTYLINE, -1, -1, -1, -1
};

static int mulearabic1[224] = {
/* 0x0020 - 0x0069 */
    ALTCHR, 0x0020, 0x0021, 0x0028, 0x0029, 0x060c, 0x06d4, 0x003a,
    0x061b, 0x061f, 0x00ab, 0x00bb, 0x000a, 0xfe80, 0xfe81, 0xfe82,
    0xfe83, 0xfe84, 0xfe85, 0xfe86, 0xfe87, 0xfe88, 0xfe8b, 0xfe8c,
    0xfe8d, 0xfe8e, 0xfe91, 0xfe92, 0xfe93, 0xfe94, 0xfe97, 0xfe98,
    0xfe9b, 0xfe9c, 0xfea9, 0xfeaa, 0xfeab, 0xfeac, 0xfead, 0xfeae,
    0xfeaf, 0xfeb0, 0xfecb, 0xfecc, 0xfecf, 0xfed0, 0xfed3, 0xfed4,
    0xfed7, 0xfed8, 0xfedf, 0xfee0, 0xfee1, 0xfee3, 0xfee4, 0xfee2,
    0xfee7, 0xfee8, 0xfee9, 0xfeeb, 0xfeec, 0xfeea, 0xfeed, 0xfeee,
    0xfef3, 0xfef4, 0xfef5, 0xfef7, 0xfef9, 0xfefb, 0xfb58, 0xfb59,
    0xfb8a, 0xfb8b, ALTCHR, ALTCHR, ALTCHR, ALTCHR, ALTCHR, ALTCHR,
/* 0x0070 - 0x00FF */
    EMPTYLINE, EMPTYLINE, EMPTYLINE, EMPTYLINE, EMPTYLINE, EMPTYLINE,
    EMPTYLINE, EMPTYLINE, EMPTYLINE
};

static int mulearabic2[224] = {
/* 0x0020 - 0x006e */
    ALTCHR, 0xfe89, 0xfe8a, 0xfe8f, 0xfe90, 0xfe95, 0xfe96, 0xfe99,
    0xfe9a, 0xfe9d, 0xfe9f, 0xfea0, 0xfe9e, 0xfea1, 0xfea3, 0xfea4,
    0xfea2, 0xfea5, 0xfea7, 0xfea8, 0xfea6, 0xfeb1, 0xfeb3, 0xfeb4,
    0xfeb2, 0xfeb5, 0xfeb7, 0xfeb8, 0xfeb6, 0xfec9, 0xfeca, 0xfecd,
    0xfeba, 0xfebd, 0xfebf, 0xfec0, 0xfebe, 0xfec1, 0xfec3, 0xfec4,
    0xfec2, 0xfec5, 0xfec7, 0xfec8, 0xfec6, 0xfec9, 0xfeca, 0xfecd,
    0xfece, 0xfed1, 0xfed2, 0xfed5, 0xfed6, ALTCHR, 0xfedb, 0xfedc,
    ALTCHR, 0xfedd, 0xfede, 0xfee5, 0xfee6, 0xfeef, 0xfef0, 0xfef1,
    0xfef2, 0xfef6, 0xfef8, 0xfefa, 0xfefc, 0xfb56, 0xfb57, 0xfb7a,
    0xfb7c, 0xfb7d, 0xfb7b, 0xfb92, 0xfb94, 0xfb95, 0xfb93, ALTCHR,
/* 0x0070 - 0x00FF */
    EMPTYLINE, EMPTYLINE, EMPTYLINE, EMPTYLINE, EMPTYLINE, EMPTYLINE,
    EMPTYLINE, EMPTYLINE, EMPTYLINE
};

static int cp437[224] = {
    0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027,
    0x0028, 0x0029, 0x002A, 0x002B, 0x002C, 0x002D, 0x002E, 0x002F,
    0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037,
    0x0038, 0x0039, 0x003A, 0x003B, 0x003C, 0x003D, 0x003E, 0x003F,
    0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047,
    0x0048, 0x0049, 0x004A, 0x004B, 0x004C, 0x004D, 0x004E, 0x004F,
    0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057,
    0x0058, 0x0059, 0x005A, 0x005B, 0x005C, 0x005D, 0x005E, 0x005F,
    0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067,
    0x0068, 0x0069, 0x006A, 0x006B, 0x006C, 0x006D, 0x006E, 0x006F,
    0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077,
    0x0078, 0x0079, 0x007A, 0x007B, 0x007C, 0x007D, 0x007E, 0x007F,
    0x00C7, 0x00FC, 0x00E9, 0x00E2, 0x00E4, 0x00E0, 0x00E5, 0x00E7,
    0x00EA, 0x00EB, 0x00E8, 0x00EF, 0x00EE, 0x00EC, 0x00C4, 0x00C5,
    0x00C9, 0x00E6, 0x00C6, 0x00F4, 0x00F6, 0x00F2, 0x00FB, 0x00F9,
    0x00FF, 0x00D6, 0x00DC, 0x00A2, 0x00A3, 0x00A5, 0x20A7, 0x0192,
    0x00E1, 0x00ED, 0x00F3, 0x00FA, 0x00F1, 0x00D1, 0x00AA, 0x00BA,
    0x00BF, 0x2310, 0x00AC, 0x00BD, 0x00BC, 0x00A1, 0x00AB, 0x00BB,
    0x2591, 0x2592, 0x2593, 0x2502, 0x2524, 0x2561, 0x2562, 0x2556,
    0x2555, 0x2563, 0x2551, 0x2557, 0x255D, 0x255C, 0x255B, 0x2510,
    0x2514, 0x2534, 0x252C, 0x251C, 0x2500, 0x253C, 0x255E, 0x255F,
    0x255A, 0x2554, 0x2569, 0x2566, 0x2560, 0x2550, 0x256C, 0x2567,
    0x2568, 0x2564, 0x2565, 0x2559, 0x2558, 0x2552, 0x2553, 0x256B,
    0x256A, 0x2518, 0x250C, 0x2588, 0x2584, 0x258C, 0x2590, 0x2580,
    0x03B1, 0x00DF, 0x0393, 0x03C0, 0x03A3, 0x03C3, 0x00B5, 0x03C4,
    0x03A6, 0x0398, 0x03A9, 0x03B4, 0x221E, 0x03C6, 0x03B5, 0x2229,
    0x2261, 0x00B1, 0x2265, 0x2264, 0x2320, 0x2321, 0x00F7, 0x2248,
    0x00B0, 0x00B7, 0x2022, 0x221A, 0x207F, 0x00B2, 0x25A0, 0x00A0
};

static int cp850[224] = {
    0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027,
    0x0028, 0x0029, 0x002A, 0x002B, 0x002C, 0x002D, 0x002E, 0x002F,
    0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037,
    0x0038, 0x0039, 0x003A, 0x003B, 0x003C, 0x003D, 0x003E, 0x003F,
    0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047,
    0x0048, 0x0049, 0x004A, 0x004B, 0x004C, 0x004D, 0x004E, 0x004F,
    0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057,
    0x0058, 0x0059, 0x005A, 0x005B, 0x005C, 0x005D, 0x005E, 0x005F,
    0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067,
    0x0068, 0x0069, 0x006A, 0x006B, 0x006C, 0x006D, 0x006E, 0x006F,
    0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077,
    0x0078, 0x0079, 0x007A, 0x007B, 0x007C, 0x007D, 0x007E, 0x007F,
    0x00C7, 0x00FC, 0x00E9, 0x00E2, 0x00E4, 0x00E0, 0x00E5, 0x00E7,
    0x00EA, 0x00EB, 0x00E8, 0x00EF, 0x00EE, 0x00EC, 0x00C4, 0x00C5,
    0x00C9, 0x00E6, 0x00C6, 0x00F4, 0x00F6, 0x00F2, 0x00FB, 0x00F9,
    0x00FF, 0x00D6, 0x00DC, 0x00F8, 0x00A3, 0x00D8, 0x20D7, 0x0192,
    0x00E1, 0x00ED, 0x00F3, 0x00FA, 0x00F1, 0x00D1, 0x00AA, 0x00BA,
    0x00BF, 0x00AE, 0x00AC, 0x00BD, 0x00BC, 0x00A1, 0x00AB, 0x00BB,
    0x2591, 0x2592, 0x2593, 0x2502, 0x2524, 0x00C1, 0x00C0, 0x00A9,
    0x2555, 0x2563, 0x2551, 0x2557, 0x255D, 0x00A2, 0x00A5, 0x2510,
    0x2514, 0x2534, 0x252C, 0x251C, 0x2500, 0x253C, 0x00E3, 0x00C3,
    0x255A, 0x2554, 0x2569, 0x2566, 0x2560, 0x2550, 0x256C, 0x00A4,
    0x00F0, 0x00D0, 0x00CA, 0x00CB, 0x00C8, 0x0131, 0x00CD, 0x00CE,
    0x00CF, 0x2518, 0x250C, 0x2588, 0x2584, 0x00A6, 0x00CC, 0x2580,
    0x00D3, 0x00DF, 0x00D4, 0x00D2, 0x00F5, 0x00D5, 0x00B5, 0x00FE,
    0x00DE, 0x00DA, 0x00DB, 0x00D9, 0x00FD, 0x00DD, 0x00AF, 0x00B4,
    0x00AD, 0x00B1, 0x2017, 0x00BE, 0x00B6, 0x00A7, 0x00F7, 0x00B8,
    0x00B0, 0x00A8, 0x00B7, 0x00B9, 0x00B3, 0x00B2, 0x25A0, 0x00A0
};

static int cp1252[224] = {
    0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027,
    0x0028, 0x0029, 0x002A, 0x002B, 0x002C, 0x002D, 0x002E, 0x002F,
    0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037,
    0x0038, 0x0039, 0x003A, 0x003B, 0x003C, 0x003D, 0x003E, 0x003F,
    0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047,
    0x0048, 0x0049, 0x004A, 0x004B, 0x004C, 0x004D, 0x004E, 0x004F,
    0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057,
    0x0058, 0x0059, 0x005A, 0x005B, 0x005C, 0x005D, 0x005E, 0x005F,
    0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067,
    0x0068, 0x0069, 0x006A, 0x006B, 0x006C, 0x006D, 0x006E, 0x006F,
    0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077,
    0x0078, 0x0079, 0x007A, 0x007B, 0x007C, 0x007D, 0x007E, 0x007F,
    0x0080, 0x0081, 0x201a, 0x0192, 0x201e, 0x2026, 0x2020, 0x2021,
    0x02c6, 0x2030, 0x0160, 0x2039, 0x0152, 0x008d, 0x008d, 0x008f,
    0x0090, 0x2018, 0x2019, 0x201c, 0x201d, 0x2022, 0x2013, 0x2014,
    0x02dc, 0x2122, 0x0161, 0x203a, 0x0153, 0x009d, 0x009e, 0x0178,
    0x00a0, 0x00a1, 0x00a2, 0x00a3, 0x00a4, 0x00a5, 0x00a6, 0x00a7,
    0x00a8, 0x00a9, 0x00aa, 0x00ab, 0x00ac, 0x00ad, 0x00ae, 0x00af,
    0x00b0, 0x00b1, 0x00b2, 0x00b3, 0x00b4, 0x00b5, 0x00b6, 0x00b7,
    0x00b8, 0x00b9, 0x00ba, 0x00bb, 0x00bc, 0x00bd, 0x00be, 0x00bf,
    0x00c0, 0x00c1, 0x00c2, 0x00c3, 0x00c4, 0x00c5, 0x00c6, 0x00c7,
    0x00c8, 0x00c9, 0x00ca, 0x00cb, 0x00cc, 0x00cd, 0x00ce, 0x00cf,
    0x00d0, 0x00d1, 0x00d2, 0x00d3, 0x00d4, 0x00d5, 0x00d6, 0x00d7,
    0x00d8, 0x00d9, 0x00da, 0x00db, 0x00dc, 0x00dd, 0x00de, 0x00df,
    0x00e0, 0x00e1, 0x00e2, 0x00e3, 0x00e4, 0x00e5, 0x00e6, 0x00e7,
    0x00e8, 0x00e9, 0x00ea, 0x00eb, 0x00ec, 0x00ed, 0x00ee, 0x00ef,
    0x00f0, 0x00f1, 0x00f2, 0x00f3, 0x00f4, 0x00f5, 0x00f6, 0x00f7,
    0x00f8, 0x00f9, 0x00fa, 0x00fb, 0x00fc, 0x00fd, 0x00fe, 0x00ff
};

int
main (int argc, char *argv[])
{
    int i;
    TT_Error fterror;
    TT_Engine engine;
    DIR *ttfdir;
    struct dirent *entry;
    struct xlfdmapping  *head = NULL;
    struct xlfdmapping  *p;

    /* first parse the command line */
    if (parse_argv (argc, argv) < 0) {
	usage (argv [0]);
	return TTMKFS_USAGE;
    }

    /* initialize the FreeType library */
    fterror = TT_Init_FreeType (&engine);

    if (fterror != TT_Err_Ok) {
	fprintf (stderr, "TT_Init_FreeType failed with %ld\n", fterror);
	return TTMKFS_FTINIT;
    }

    ttfdir = opendir (path);

    if (ttfdir == NULL) {
	fprintf (stderr, "Could not read directory %s: %s\n", path, strerror (errno));
        return TTMKFS_DIR;
    }

    while ((entry = readdir (ttfdir)) != NULL) {
	TT_Face face;
	TT_Face_Properties  properties;
	char *filename;
	const char *font_name;
	const char *foundry_name;
	const char *weight_name;
	const char *widht_name;

	if (!is_ttf (entry)) {
	    continue;
	}

	filename = malloc (strlen (path) + strlen (entry->d_name) + 2);

	if (filename == NULL) {
	    return TTMKFS_MEM;
	}

	/* assemble the complete file name of the TTF file */
	strcpy (filename, path);
	strcat (filename, "/");
	strcat (filename, entry->d_name);

	/* open the TTF file */
	fterror = TT_Open_Face (engine, filename, &face);
	
	if (fterror != TT_Err_Ok) {
	    fprintf (stderr, "TT_Open_Face failed with %ld for file %s\n", fterror, filename);
	    free (filename);
	    continue;
	}

	/* ask FreeType for the font properties */
	fterror = TT_Get_Face_Properties (face, &properties);

	if (fterror != TT_Err_Ok) {
	    fprintf (stderr, "TT_Get_Face_Properties failed with %ld for file %s\n", fterror, filename);
	    goto bail;
	}

	font_name    = get_name (face);
	foundry_name = get_foundry (properties.os2->achVendID);
	weight_name  = get_weight (filename, &properties);
	widht_name   = get_width_name (filename, &properties);

	count_encodings = 0;
	
	for (i = 0; i < properties.num_CharMaps; i++) {
	    TT_UShort platform;
	    TT_UShort encoding;
	    
	    TT_Get_CharMap_ID (face, i, &platform, &encoding);

	    if (platform == TT_PLATFORM_MICROSOFT) {
		switch (encoding) {
		case TT_MS_ID_UNICODE_CS:
		    {
			TT_CharMap cmap;
			struct {
			    int *table;
			    char *encoding;
			    enum completeness completeness;
				int  enc_weight;
			} *p, char_mappings [] = {
			    {iso8859_1,  "iso8859-1",         IGNORE, 0},
			    {iso8859_2,  "iso8859-2",         IGNORE, 1},
			    {iso8859_3,  "iso8859-3",         IGNORE, 2},
			    {iso8859_4,  "iso8859-4",         IGNORE, 2},
			    {iso8859_5,  "iso8859-5",         IGNORE, 1},
			    {iso8859_6,  "iso8859-6",         IGNORE, 3},
			    {iso8859_7,  "iso8859-7",         IGNORE, 2},
			    {iso8859_8,  "iso8859-8",         IGNORE, 3},
			    {iso8859_9,  "iso8859-9",         IGNORE, 1},
			    {iso8859_10, "iso8859-10",        IGNORE, 1},
			    {iso8859_11, "tis620.2533-1",     IGNORE, 0},
			    {iso8859_11, "tis620-0",          IGNORE, 4},
			    {tis620_2,   "tis620-2",          IGNORE, 0},
			    {iso8859_13, "iso8859-13",        IGNORE, 3},
			    {iso8859_14, "iso8859-14",        IGNORE, 3},
			    {iso8859_15, "iso8859-15",        IGNORE, 0},
			    {iso8859_16, "iso8859-16",        IGNORE, 3},
	            {iso8859_9e, "iso8859-9e",        IGNORE, 4},
/*			    {koi8r,      "koi8-r",            STRICT, 0},
			    {koi8ru,     "koi8-u",            STRICT, 0}, */
			    {koi8r1,     "koi8-r",            IGNORE, 0},
			    {koi8ru1,    "koi8-u",            IGNORE, 0},
			    {koi8k,      "koi8-k",            IGNORE, 4},
			    {koi8t,      "koi8-t",            IGNORE, 3},
			    {mscp1251,	 "microsoft-cp1251",  STRICT, 1},
			    {mscp1251tt, "tatar-cyr",         STRICT, 3},
			    {mscp1255,	 "microsoft-cp1255",  STRICT, 1},
			    {mscp1256,	 "microsoft-cp1256",  STRICT, 1},
			    {jisx0201,   "jisx0201.1976-0",   STRICT, 1},
			    {jisx0208,   "jisx0208.1983-0",   STRICT, 2},
			    {jisx0212,   "jisx0212.1990-0",   STRICT, 3},
			    {gb2312,     "gb2312.1980-0",     STRICT, 2},
			    {gbk,        "gbk-0",             STRICT, 2},
			    {big5,       "big5-0",	          STRICT, 2},
			    {ksc5601,    "ksc5601.1987-0",    STRICT, 2},
			    {johab,      "ksc5601.1992-3",    STRICT, 2},
			    {viscii,     "viscii1.1-1",       STRICT, 1},
			    {tcvn,       "tcvn-5712",         STRICT, 4},
			    {armscii,    "armscii-8",         STRICT, 3},
			    {georgianrs, "georgian-academy",  STRICT, 4},
			    {georgianstd,"georgian-ps",       STRICT, 0},
			    {mulelao,    "mulelao-1",         STRICT, 3},
			    {asciilao,   "ibm-cp1133",        STRICT, 0},
			    {mulearabic0,"mulearabic-0",      STRICT, 0},
			    {mulearabic1,"mulearabic-1",      STRICT, 0},
			    {mulearabic2,"mulearabic-2",      STRICT, 3},
			    {urdunaqsh0, "urdunaqsh-0",       STRICT, 4},
			    {cp437,      "dosencoding-cp437", STRICT, 0},
			    {cp850,      "dosencoding-cp850", STRICT, 0},
			    {A_Z_letters,"ascii-0",           LOOSE,  0},
			    {NULL,       NULL,                IGNORE, 0}
			};

			TT_Get_CharMap (face, i, &cmap);

			for (p = char_mappings; p->table; p++) {
			    if ((p->completeness == IGNORE) ||
				(p->completeness == completeness))
			    {
				if (has_8bit_encoding (&cmap, p->table)) {
				    add_mapping (&head, entry->d_name, foundry_name,
						 font_name, weight_name, widht_name,
						 &properties, p->encoding);
					count_encodings += p->enc_weight;
				}
			    }
			}
			if (always_unicode) {
		    	add_mapping (&head, entry->d_name, foundry_name,
					 font_name, weight_name, widht_name,
					 &properties, "iso10646-1");
			} else if (count_encodings >= COUNT_ENCODINGS_UNICODE) {
		    	add_mapping (&head, entry->d_name, foundry_name,
					 font_name, weight_name, widht_name,
					 &properties, "iso10646-1");
			} else if (has_8bit_encoding (&cmap, ethiopic)) {
		    	add_mapping (&head, entry->d_name, foundry_name,
					 font_name, weight_name, widht_name,
					 &properties, "iso10646-1");
			} else if (has_8bit_encoding (&cmap, inuit)) {
		    	add_mapping (&head, entry->d_name, foundry_name,
					 font_name, weight_name, widht_name,
					 &properties, "iso10646-1");
			} else if (has_8bit_encoding (&cmap, iso8859_9e)) {
		    	add_mapping (&head, entry->d_name, foundry_name,
					 font_name, weight_name, widht_name,
					 &properties, "iso10646-1");
			} else if (has_8bit_encoding (&cmap, georgianstd)) {
		    	add_mapping (&head, entry->d_name, foundry_name,
					 font_name, weight_name, widht_name,
					 &properties, "iso10646-1");
			} else if (has_8bit_encoding (&cmap, koi8t)) {
		    	add_mapping (&head, entry->d_name, foundry_name,
					 font_name, weight_name, widht_name,
					 &properties, "iso10646-1");
			} else if (has_8bit_encoding (&cmap, armscii)) {
		    	add_mapping (&head, entry->d_name, foundry_name,
					 font_name, weight_name, widht_name,
					 &properties, "iso10646-1");
			} else if (has_8bit_encoding (&cmap, mulelao)) {
		    	add_mapping (&head, entry->d_name, foundry_name,
					 font_name, weight_name, widht_name,
					 &properties, "iso10646-1");
			}
			}
		    break;
		case TT_MS_ID_SJIS:
		    add_mapping (&head, entry->d_name, foundry_name,
				 font_name, weight_name, widht_name,
				 &properties, "jisx0208.1983-0");
		    add_mapping (&head, entry->d_name, foundry_name,
				 font_name, weight_name, widht_name,
				 &properties, "jisx0201.1976-0");
			if (always_unicode)
					add_mapping (&head, entry->d_name, foundry_name,
						font_name, weight_name, widht_name,
						&properties, "iso10646-1");
		    break;
		case TT_MS_ID_GB2312:
		    add_mapping (&head, entry->d_name, foundry_name,
				 font_name, weight_name, widht_name,
				 &properties, "gb2312.1980-0");
			if (always_unicode)
					add_mapping (&head, entry->d_name, foundry_name,
						font_name, weight_name, widht_name,
						&properties, "iso10646-1");
		    break;
		case TT_MS_ID_BIG_5:
		    add_mapping (&head, entry->d_name, foundry_name,
				 font_name, weight_name, widht_name,
				 &properties, "big5-0");
			if (always_unicode)
					add_mapping (&head, entry->d_name, foundry_name,
						font_name, weight_name, widht_name,
						&properties, "iso10646-1");
		    break;
		case TT_MS_ID_WANSUNG:
		    add_mapping (&head, entry->d_name, foundry_name,
				 font_name, weight_name, widht_name,
				 &properties, "ksc5601.1987-0");
			if (always_unicode)
					add_mapping (&head, entry->d_name, foundry_name,
						font_name, weight_name, widht_name,
						&properties, "iso10646-1");
		    break;
		case TT_MS_ID_JOHAB:
		    add_mapping (&head, entry->d_name, foundry_name,
				 font_name, weight_name, widht_name,
				 &properties, "ksc5601.1992-3");
			if (always_unicode)
					add_mapping (&head, entry->d_name, foundry_name,
						font_name, weight_name, widht_name,
						&properties, "iso10646-1");
		    break;
		case TT_MS_ID_SYMBOL_CS:
		    {
			TT_CharMap cmap;

			TT_Get_CharMap (face, i, &cmap);
			if (has_8bit_encoding (&cmap, cp1252)) {
			    add_mapping (&head, entry->d_name, foundry_name,
					 font_name, weight_name, widht_name,
					 &properties, "misc-fontspecific");
			} else {
			    add_mapping (&head, entry->d_name, foundry_name,
					 font_name, weight_name, widht_name,
					 &properties, "microsoft-symbol");
			}
			break;
		    }
		}
	    }
	    if (platform == TT_PLATFORM_MACINTOSH) {
		TT_CharMap cmap;

		switch (encoding) {
		case TT_MAC_ID_ROMAN:
		    {
			struct {
			    int *table;
			    char *encoding;
			    enum completeness completeness;
			} *p, char_mappings [] = {
			/*  {iso8859_1,  "iso8859-1",  IGNORE}, often this is wrong */
			    {iso8859_2,  "iso8859-2",  IGNORE},
			    {iso8859_3,  "iso8859-3",  IGNORE},
			    {iso8859_4,  "iso8859-4",  IGNORE},
			    {iso8859_5,  "iso8859-5",  IGNORE},
			    {iso8859_6,  "iso8859-6",  IGNORE},
			    {iso8859_7,  "iso8859-7",  IGNORE},
			    {iso8859_8,  "iso8859-8",  IGNORE},
			    {iso8859_9,  "iso8859-9",  IGNORE},
			    {iso8859_10, "iso8859-10", IGNORE},
			    {iso8859_11, "tis620.2533-1", IGNORE},
			    {iso8859_11, "tis620-0", IGNORE},
			    {tis620_2,   "tis620-2", IGNORE},
			    {iso8859_13, "iso8859-13", IGNORE},
			    {iso8859_14, "iso8859-14", IGNORE},
			    {iso8859_15, "iso8859-15", IGNORE},
			    {iso8859_9e, "iso8859-9e", IGNORE},
			    {koi8r,      "koi8-r",     STRICT},
			    {koi8ru,     "koi8-u",     STRICT},
			    {koi8r1,     "koi8-r",     LOOSE},
			    {koi8ru1,    "koi8-u",     LOOSE},
			    {koi8k,      "koi8-k",     LOOSE},
			    {koi8t,      "koi8-t",     LOOSE},
			    {mscp1251,	 "microsoft-cp1251", STRICT},
			    {mscp1251tt, "tatar-cyr",  STRICT},
			    {mscp1255,	 "microsoft-cp1255", STRICT},
			    {mscp1256,	 "microsoft-cp1256", STRICT},
			    {jisx0201,   "jisx0201.1976-0", STRICT},
			    {viscii,     "viscii1.1-1", STRICT},
			    {tcvn,       "tcvn-5712",  STRICT}, 
			    {armscii,    "armscii-8",  STRICT},
			    {georgianrs, "georgian-academy", STRICT},
			    {georgianstd,"georgian-ps", STRICT},
			    {mulelao,    "mulelao-1",  STRICT},
			    {asciilao,   "ibm-cp1133", STRICT},
			    {mulearabic0,"mulearabic-0", STRICT},
			    {mulearabic1,"mulearabic-1", STRICT},
			    {mulearabic2,"mulearabic-2", STRICT},
			    {urdunaqsh0, "urdunaqsh-0", STRICT},
			    {cp437, "dosencoding-cp437", STRICT},
			    {cp850, "dosencoding-cp850", STRICT},
			    {A_Z_letters,"ascii-0",    LOOSE},
			    {NULL,       NULL,         IGNORE}
			};

			TT_Get_CharMap (face, i, &cmap);
			for (p = char_mappings; p->table; p++) {
			    if ((p->completeness == IGNORE) ||
			        (p->completeness == completeness))
			    {
				if (has_8bit_encoding (&cmap, p->table)) {
				    add_mapping (&head, entry->d_name, foundry_name,
						 font_name, weight_name, widht_name,
						 &properties, p->encoding);
				}
			    }
			}
		    }
		    break;
		case TT_MAC_ID_JAPANESE:
		    add_mapping (&head, entry->d_name, foundry_name,
				 font_name, weight_name, widht_name,
				 &properties, "jisx0208.1983-0");
		    add_mapping (&head, entry->d_name, foundry_name,
				 font_name, weight_name, widht_name,
				 &properties, "jisx0201.1976-0");
		    break;
		case TT_MAC_ID_TRADITIONAL_CHINESE:
		    add_mapping (&head, entry->d_name, foundry_name,
				 font_name, weight_name, widht_name,
				 &properties, "big5-0");
		    break;
		case TT_MAC_ID_KOREAN:
		    add_mapping (&head, entry->d_name, foundry_name,
				 font_name, weight_name, widht_name,
				 &properties, "ksc5601.1987-0");
		    break;
		case TT_MAC_ID_ARABIC:
		    TT_Get_CharMap (face, i, &cmap);
		    if (has_8bit_encoding (&cmap, iso8859_6)) {
			add_mapping (&head, entry->d_name, foundry_name,
				     font_name, weight_name, widht_name,
				     &properties, "iso8859-6");
		    }
		    break;
		case TT_MAC_ID_HEBREW:
		    TT_Get_CharMap (face, i, &cmap);
		    if (has_8bit_encoding (&cmap, iso8859_8)) {
			add_mapping (&head, entry->d_name, foundry_name,
				     font_name, weight_name, widht_name,
				     &properties, "iso8859-8");
		    }
		    break;
		case TT_MAC_ID_GREEK:
		    TT_Get_CharMap (face, i, &cmap);
		    if (has_8bit_encoding (&cmap, iso8859_7)) {
			add_mapping (&head, entry->d_name, foundry_name,
				     font_name, weight_name, widht_name,
				     &properties, "iso8859-7");
		    }
		    break;
		case TT_MAC_ID_RUSSIAN:
		    {
			struct {
			    int *table;
			    char *encoding;
			    enum completeness completeness;
			} *p, char_mappings [] = {
			    {iso8859_5,  "iso8859-5",  IGNORE},
			    {koi8r,      "koi8-r",     STRICT},
			    {koi8ru,     "koi8-u",     STRICT},
			    {koi8r1,     "koi8-r",     LOOSE},
			    {koi8ru1,    "koi8-u",     LOOSE},
			    {mscp1251,	 "microsoft-cp1251", STRICT},
			    {NULL,       NULL,         IGNORE}
			};

			TT_Get_CharMap (face, i, &cmap);
			for (p = char_mappings; p->table; p++) {
			    if ((p->completeness == IGNORE) ||
			        (p->completeness == completeness))
			    {
				if (has_8bit_encoding (&cmap, p->table)) {
				    add_mapping (&head, entry->d_name, foundry_name,
						 font_name, weight_name, widht_name,
						 &properties, p->encoding);
				}
			    }
			}
		    }
		    break;
//case TT_MAC_ID_RSYMBOL:
		case TT_MAC_ID_DEVANAGARI:
		    add_mapping (&head, entry->d_name, foundry_name,
				 font_name, weight_name, widht_name,
				 &properties, "is13194-devanagari");
		    break;
//case TT_MAC_ID_GURMUKHI:
//case TT_MAC_ID_GUJARATI:
//case TT_MAC_ID_ORIYA:
//case TT_MAC_ID_BENGALI:
//case TT_MAC_ID_TAMIL:
//case TT_MAC_ID_TELUGU:
//case TT_MAC_ID_KANNADA:
//case TT_MAC_ID_MALAYALAM:
//case TT_MAC_ID_SINHALESE:
//case TT_MAC_ID_BURMESE:
//case TT_MAC_ID_KHMER:
		case TT_MAC_ID_THAI:
		    TT_Get_CharMap (face, i, &cmap);
		    if (has_8bit_encoding (&cmap, iso8859_11)) {
			add_mapping (&head, entry->d_name, foundry_name,
				     font_name, weight_name, widht_name,
				     &properties, "tis620.2533-1");
			add_mapping (&head, entry->d_name, foundry_name,
				     font_name, weight_name, widht_name,
				     &properties, "tis620-0");
			add_mapping (&head, entry->d_name, foundry_name,
				     font_name, weight_name, widht_name,
				     &properties, "tis620-1");
		    }
		    break;
		case TT_MAC_ID_LAOTIAN:
		    if (has_8bit_encoding (&cmap, mulelao)) {
		    add_mapping (&head, entry->d_name, foundry_name,
				 font_name, weight_name, widht_name,
				 &properties, "mulelao-1");
		    }
                    if (has_8bit_encoding (&cmap, asciilao)) {
                    add_mapping (&head, entry->d_name, foundry_name,
                                 font_name, weight_name, widht_name,
                                 &properties, "ibm-cp1133");
                    }
		    break;
		case TT_MAC_ID_GEORGIAN:
		    if (has_8bit_encoding (&cmap, georgianrs)) {
                        add_mapping (&head, entry->d_name, foundry_name,
                                     font_name, weight_name, widht_name,
                                     &properties, "georgian-acedemy");
		    }
		    if (has_8bit_encoding (&cmap, georgianstd)) {
                        add_mapping (&head, entry->d_name, foundry_name,
                                     font_name, weight_name, widht_name,
				     &properties, "georgian-ps");
		    }
		    break;
		case TT_MAC_ID_ARMENIAN:
                    if (has_8bit_encoding (&cmap, armscii)) {
                        add_mapping (&head, entry->d_name, foundry_name,
                                     font_name, weight_name, widht_name,
                                     &properties, "armscii-8");
                    }
		    break;
//case TT_MAC_ID_MALDIVIAN:
		case TT_MAC_ID_SIMPLIFIED_CHINESE:
		    add_mapping (&head, entry->d_name, foundry_name,
				 font_name, weight_name, widht_name,
				 &properties, "gb2312.1980-0");
		    break;
		case TT_MAC_ID_TIBETAN:
		    add_mapping (&head, entry->d_name, foundry_name,
				 font_name, weight_name, widht_name,
				 &properties, "muletibetan-0");
		    break;
//case TT_MAC_ID_MONGOLIAN:
//case TT_MAC_ID_GEEZ:
		case TT_MAC_ID_SLAVIC:
		    {
			struct {
			    int *table;
			    char *encoding;
			    enum completeness completeness;
			} *p, char_mappings [] = {
			    {iso8859_2,  "iso8859-2",  STRICT},
			    {iso8859_3,  "iso8859-3",  IGNORE},
			    {iso8859_4,  "iso8859-4",  IGNORE},
			    {iso8859_9,  "iso8859-9",  IGNORE},
			    {iso8859_10, "iso8859-10", IGNORE},
			    {iso8859_13, "iso8859-13", IGNORE},
			    {NULL,       NULL,         IGNORE}
			};

			TT_Get_CharMap (face, i, &cmap);
			for (p = char_mappings; p->table; p++) {
			    if ((p->completeness == IGNORE) ||
			        (p->completeness == completeness))
			    {
				if (has_8bit_encoding (&cmap, p->table)) {
				    add_mapping (&head, entry->d_name, foundry_name,
						 font_name, weight_name, widht_name,
						 &properties, p->encoding);
				}
			    }
			}
		    }
		    break;
		case TT_MAC_ID_VIETNAMESE:
		    TT_Get_CharMap (face, i, &cmap);
		    if (has_8bit_encoding (&cmap, viscii)) {
			add_mapping (&head, entry->d_name, foundry_name,
				     font_name, weight_name, widht_name,
				     &properties, "viscii1.1-1");
		    }
                    if (has_8bit_encoding (&cmap, tcvn)) {
                        add_mapping (&head, entry->d_name, foundry_name,
                                     font_name, weight_name, widht_name,
                                     &properties, "tcvn-5712");
                    }
		    break;
//case TT_MAC_ID_SINDHI:
//case TT_MAC_ID_UNINTERP:

		}
	    }
	}

	if (create_debug_entries) {
	    add_mapping (&head, entry->d_name, foundry_name,
			 font_name, weight_name, widht_name,
			 &properties, "truetype-raw");
	}
	
    bail:
	/* we are done with this file */
	TT_Close_Face (face);
	free (filename);
    }

    closedir (ttfdir);

    /* we have processed all the files */
    /* now write the number of mappings we actually got */
    fprintf (outstream, "%lu\n", entries);

    /* and append our collected mappings */
    for (p = head; p; p = p->next) {
	fputs (p->mapping, outstream);
    }

    /* shut down */
    TT_Done_FreeType (engine);
    fclose (outstream);

    return 0;
}

int
parse_argv (int argc, char *argv[])
{
    int c;

    while ((c = getopt (argc, argv, "d:f:o:m:ucpg")) != -1) {
	switch (c) {
	case 'o':
	    outstream = fopen (optarg, "w");

	    if (outstream == NULL) {
		fprintf (stderr, "Could not open file %s: %s\n", optarg, strerror (errno));
		return -1;
	    }
	    break;
	case 'u':
		always_unicode = 1;
		break;
	case 'd':
	    path = strdup (optarg);
	    break;
	case 'f':
	    default_foundry = strdup (optarg);
	    break;
	case 'm':
	    max_missing = atoi (optarg);
	    break;
	case 'c':
	    completeness = LOOSE;
	    break;
	case 'p':
	    use_panose = 1;
	    break;
	case 'g':
	    create_debug_entries = 1;
	    break;
	default:
	case '?':
	case ':':
	    return -1;
	    break;
	}
    }

    if (path == NULL)
	path = strdup (DEFAULTDIR);

    if (default_foundry == NULL)
	default_foundry = strdup (DEFAULTFOUNDRY);

    if (outstream == 0) {
	outstream = stdout;
    }

    return 0;
}

void
usage (const char *name)
{
    fprintf (stderr, "This Program is (C) Joerg Pommnitz, 1998\n");
    fprintf (stderr, "Usage: %s [-u] [-d dir] [-o filename] [-f foundry] [-m #]\n", name);
    fprintf (stderr, "   -u always output unicode (*-iso10646-1) font lines\n");
	fprintf (stderr, "   -d scan directory dir for TrueType files (def. is current directory)\n");
    fprintf (stderr, "   -o write output to file named filename (def. stdout)\n");
    fprintf (stderr, "   -f use foundry for unknown font vendors (def. "DEFAULTFOUNDRY")\n");
    fprintf (stderr, "   -m max # of missing characters per encoding (def. 5)\n");
    fprintf (stderr, "   -c use less strict completeness tests for encoding tables\n");
    fprintf (stderr, "   -p use panose information\n");
}

int
is_ttf (const struct dirent *e)
{
    int len;

    /* we make the decision by the extension of the file name */
    return ((len = strlen (e->d_name)) > 4) &&
	(e->d_name[len - 4] == '.') &&
	(toupper(e->d_name[len - 3]) == 'T') &&
	(toupper(e->d_name[len - 2]) == 'T') &&
	(toupper(e->d_name[len - 1]) == 'F');
}

const char *
get_foundry (const TT_Char orig_code[4])
{

    char code[5];
    
    static struct {
	const char  code[4];	/* TTF foundry code */
	const char *xlfd;	/* xlfd foundry name */
    } *p, known_foundries[] = {
	{ "2REB", "2Rebels"},
	{ "3ip" , "Three Islands Press"},
	{ "3ip ", "Three Islands Press"},
	{ "ABC" , "Altek Instruments"},
	{ "ABC ", "Altek Instruments"},
	{ "ACG" , "Agfa"},
	{ "ACUT", "Acute Type"},
	{ "ADBE", "adobe"},
	{ "AGFA", "Agfa Typographic Systems"},
	{ "ALTS", "altsys"},
	{ "AOP" , "An Art of Penguin"},
	{ "AOP ", "An Art of Penguin"},
	{ "APPL", "Apple"},
	{ "ARPH", "Arphic Technology Co."},
	{ "ATEC", "alltype"},
	{ "AZLS", "Azalea Software, Inc."},
	{ "B&H",  "b&h"}, /* Bigelow & Holmes" */
	{ "B&H ", "b&h"}, /* Bigelow & Holmes" */
	{ "BARS", "CIA (BAR CODES) UK"},
	{ "BERT", "Berthold"},
	{ "BITM", "Bitmap Software"},
	{ "BITS", "Bitstream"},
	{ "BLAH", "Mister Bla's Fontworx"},
	{ "BOYB", "I. Frances"},
	{ "BRTC", "Bear Rock Technologies"},
	{ "BWFW", "B/W Fontworks"},
	{ "C&C",  "Carter & Cone"},
	{ "C&C ", "Carter & Cone"},
	{ "CAK" , "pluginfonts.com"},
	{ "CAK ", "pluginfonts.com"},
	{ "CANO", "cannon"},
	{ "CASL", "H.W. Caslon & Company Ltd."},
	{ "COOL", "Cool Fonts"},
	{ "CTDL", "China Type Design Ltd."},
	{ "DAMA", "Dalton Maag Limited"},
	{ "DS"  , "Dainippon Screen Mfg. Co., Inc."},
	{ "DS  ", "Dainippon Screen Mfg. Co., Inc."},
	{ "DSCI", "Design Science, Inc."},
	{ "DTC",  "Digital Typeface Corp."},
	{ "DTC ", "Digital Typeface Corp."},
	{ "DTPS", "DTP Software"},
	{ "DUXB", "Duxbury Systems, Inc."},
	{ "DYNA", "DynaLab"},
	{ "ECF" , "emerald city fontwerks"},
	{ "ECF ", "emerald city fontwerks"},
	{ "EDGE", "Rivers Edge Corp."},
	{ "EFF" , "Electronic Font Foundry"},
	{ "EFF ", "Electronic Font Foundry"},
	{ "EFNT", "E Fonts L.L.C."},
	{ "ELSE", "elseware"},
	{ "EPSN", "epson"},
	{ "ERAM", "eraman"},
	{ "ESIG", "E-Signature"},
	{ "FBI",  "The Font Bureau, Inc."},
	{ "FBI ", "The Font Bureau, Inc."},
	{ "FCAB", "The Font Cabinet"},
	{ "FJ",   "Fujitsu"},
	{ "FJ  ", "Fujitsu"},
	{ "FONT", "Font Source"},
	{ "FS"  , "Formula Solutions"},
	{ "FS  ", "Formula Solutions"},
	{ "FSE" , "Font Source Europe"},
	{ "FSE ", "Font Source Europe"},
	{ "FSI" , "FSI Fonts und Software GmbH"},
	{ "FSI ", "FSI Fonts und Software GmbH"},
	{ "FTFT", "FontFont"},
	{ "FWRE", "Fontware Limited"},
	{ "GALA", "Galapagos Design Group, Inc."},
	{ "GD"  , "GD Fonts"},
	{ "GD  ", "GD Fonts"},
	{ "GLYF", "Glyph Systems"},
	{ "GPI",  "Gamma Productions, Inc."},
	{ "GPI ", "Gamma Productions, Inc."},
	{ "HILL", "Hill Systems"},
	{ "HOUS", "House Industries"},
	{ "HP",   "hp"},
	{ "HP  ", "hp"},
	{ "HY",   "HanYang System"},
	{ "HY  ", "HanYang System"},
	{ "IBM",  "IBM"},
	{ "IBM ", "IBM"},
	{ "IDEE", "IDEE TYPOGRAFICA"},
	{ "IDF",  "International Digital Fonts"},
	{ "IDF ", "International Digital Fonts"},
	{ "ILP",  "Indigenous Languages Project"},
	{ "ILP ", "Indigenous Languages Project"},
	{ "ITC",  "itc"},
	{ "ITC ", "itc"},
	{ "IMPR", "impress"},
	{ "ITF" , "International Typefounders, Inc."},
	{ "ITF ", "International Typefounders, Inc."},
	{ "KATF", "Kingsley/ATF"},
	{ "LANS", "Lanston Type Co., Ltd."},
	{ "LARA", "Larabiefonts"},
	{ "LEAF", "Interleaf, Inc."},
	{ "LETR", "letraset"},
	{ "LGX" , "Logix Research Institute, Inc."},
	{ "LGX ", "Logix Research Institute, Inc."},
	{ "LING", "Linguist's Software"},
	{ "LINO", "linotype"},
	{ "LP",   "LetterPerfect Fonts"},
	{ "LP  ", "LetterPerfect Fonts"},
	{ "LTRX", "lighttracks"},
	{ "MACR", "macromedia"},
	{ "MC"  , "Cerajewski Computer Consulting"},
	{ "MC  ", "Cerajewski Computer Consulting"},
	{ "MILL", "Millan"},
	{ "MJ"  , "Majus Corporation"},
	{ "MJ  ", "Majus Corporation"},
	{ "MLGC", "Micrologic Software"},
	{ "MONO", "monotype"},
	{ "MS",   "microsoft"},
	{ "MS  ", "microsoft"},
	{ "MSCR", "Majus Corporation"},
	{ "MT",   "monotype"},
	{ "MT  ", "monotype"},
	{ "MTY" , "Motoya Co. ,LTD."},
	{ "MTY ", "Motoya Co. ,LTD."},
	{ "MUTF", "CACHE Enterprise Sdn. Bhd"},
	{ "NB"  , "No Bodoni Typography"},
	{ "NB  ", "No Bodoni Typography"},
	{ "NDTC", "Neufville Digital"},
	{ "NEC",  "nec"},
	{ "NEC ", "nec"},
	{ "NIS" , "NIS Corporation"},
	{ "NIS ", "NIS Corporation"},
	{ "ORBI", "Orbit Enterprises, Inc."},
	{ "P22" , "P22 Inc."},
	{ "P22 ", "P22 Inc."},
	{ "PARA", "ParaType"},
	{ "PDWX", "Parsons Design Workx"},
	{ "PF"  , "Phil's Fonts, Inc."},
	{ "PF  ", "Phil's Fonts, Inc."},
	{ "PRFS", "Production First Software"},
	{ "QMSI", "QMS/Imagen"},
	{ "RICO", "Ricoh"},
	{ "RKFN", "R K Fonts"},
	{ "robo", "Buro Petr van Blokland"},
	{ "RUDY", "RudynFluffy"},
	{ "SAX" , "s.a.x. Software gmbh"},
	{ "SAX ", "s.a.x. Software gmbh"},
	{ "Sean", "The FontSite"},
	{ "SFI" , "Software Friends"},
	{ "SFI ", "Software Friends"},
	{ "SFUN", "Soft Union"},
	{ "SG"  , "Scooter Graphics"},
	{ "SG  ", "Scooter Graphics"},
	{ "SIG" , "Signature Software, Inc."},
	{ "SIG ", "Signature Software, Inc."},
	{ "skz" , "Celtic Lady's Fonts"},
	{ "skz ", "Celtic Lady's Fonts"},
	{ "SN"  , "SourceNet"},
	{ "SN  ", "SourceNet"},
	{ "SOHO", "Soft Horizons"},
	{ "SOS" , "Standing Ovations Software"},
	{ "SOS ", "Standing Ovations Software"},
	{ "STF" , "Brian Sooy & Co + Sooy Type Foundry"},
	{ "STF ", "Brian Sooy & Co + Sooy Type Foundry"},
	{ "STON", "ZHUHAI Stone"},
	{ "SUNW", "sunwalk fontworks"},
	{ "SWFT", "Swfte International"},
	{ "SYN" , "SynFonts"},
	{ "SYN ", "SynFonts"},
	{ "TDR" , "Tansin A. Darcos & Co."},
	{ "TDR ", "Tansin A. Darcos & Co."},
	{ "TF"  , "Treacyfaces/Headliners"},
	{ "TF  ", "Treacyfaces/Headliners"},
	{ "TILD", "SIA Tilde"},
	{ "TPTC", "TEST PILOT"},
	{ "TR"  , "Type Revivals"},
	{ "TR  ", "Type Revivals"},
	{ "TS"  , "TamilSoft Corporation"},
	{ "TS  ", "TamilSoft Corporation"},
	{ "UA"  , "UnAuthorized Type"},
	{ "UA  ", "UnAuthorized Type"},
	{ "UNKN", DEFAULTFOUNDRY},
	{ "URW",  "urw"},
	{ "URW ", "urw"},
	{ "VKP" , "Vijay K. Patel"},
	{ "VKP ", "Vijay K. Patel"},
	{ "VLKF", "Visualogik Technology & Design"},
	{ "VOG" , "Martin Vogel"},
	{ "VOG ", "Martin Vogel"},
	{ "Y&Y" , "y&y"},
	{ "Y&Y ", "y&y"},
	{ "ZeGr", "Zebra Font Factory"},
	{ "zeta", "Tangram Studio"},
	{ "ZSFT", "ZSoft"},
	{ {0,},   NULL}
    };

    code[0] = toupper(orig_code[0]);
    code[1] = toupper(orig_code[1]);
    code[2] = toupper(orig_code[2]);
    code[3] = toupper(orig_code[3]);
    code[4] = 0;

    for (p = known_foundries; p->xlfd; p++) {
	if (memcmp (p->code, code, strlen (code)) == 0) {
	    return p->xlfd;
	}
    }

    fprintf (stderr, "unknown font foundry code %s\n", code);
    return default_foundry;
}

static const char *
do_get_name (TT_Face face, int idx, int unicode)
{
    char*        string;
    TT_UShort    string_len;
    static char  name_buffer[257];
    int          i;
    int          name_len = 0;
    int          increment;
    
    TT_Get_Name_String (face, idx, &string, &string_len);
    
    if (unicode) {
	string_len =  string_len < 512 ? string_len : 512;
	increment = 2;
	i = 1;
    } else {
	string_len =  string_len < 257 ? string_len : 257;
	increment = 1;
	i = 0;
    }

    for (; i < string_len; i += increment) {
	if (!isascii (string[i])) {
	    return NULL;
	}
	name_buffer[name_len++] = (string[i] == '-') ? '_' : string[i];
    }

    name_buffer[name_len] = '\0';

    return name_buffer;
}

static int
isMacEnglish (TT_UShort  platform, TT_UShort  encoding, TT_UShort  language)
{
    return ((platform == TT_PLATFORM_MACINTOSH) &&
	    (encoding == TT_MAC_ID_ROMAN) &&
	    (language == TT_MAC_LANGID_ENGLISH));
}

static int
isUnicodeEnglish (TT_UShort  platform, TT_UShort  encoding, TT_UShort  language)
{
    if ((platform == TT_PLATFORM_MICROSOFT) &&
	((encoding == TT_MS_ID_UNICODE_CS) || (encoding == TT_MS_ID_SYMBOL_CS))) {
	return  ((language == TT_MS_LANGID_ENGLISH_UNITED_STATES) ||
		 (language == TT_MS_LANGID_ENGLISH_UNITED_KINGDOM) ||
		 (language == TT_MS_LANGID_ENGLISH_AUSTRALIA) ||
		 (language == TT_MS_LANGID_ENGLISH_CANADA) ||
		 (language == TT_MS_LANGID_ENGLISH_NEW_ZEALAND) ||
		 (language == TT_MS_LANGID_ENGLISH_IRELAND) ||
		 (language == TT_MS_LANGID_ENGLISH_SOUTH_AFRICA) ||
		 (language == TT_MS_LANGID_ENGLISH_JAMAICA) ||
		 (language == TT_MS_LANGID_ENGLISH_CARIBBEAN) ||
		 (language == TT_MS_LANGID_ENGLISH_BELIZE) ||
		 (language == TT_MS_LANGID_ENGLISH_TRINIDAD) ||
		 (language == TT_MS_LANGID_ENGLISH_ZIMBABWE) ||
		 (language == TT_MS_LANGID_ENGLISH_PHILIPPINES));
    } else if (platform == TT_PLATFORM_APPLE_UNICODE) {
	return (1); /* Apple Unicode has no language identifiers */
    }

    return 0;
}

/*
 * This function is cut&paste from the FreeType-1.0 sample program
 * ftdump with minor editing.
 */
const char *
get_name (TT_Face face)
{
    int i, n;
    const char *result = NULL;
    const char *ps_name = NULL;

    TT_UShort  platform;
    TT_UShort  encoding;
    TT_UShort  language;
    TT_UShort  NameId;


    /* get the number of name strings in this font */
    if ((n = TT_Get_Name_Count (face)) < 0) {
	return "unknown";
    }

    /* now look at every entry */
    for (i = 0; i < n; i++) {
	TT_Get_Name_ID (face, i, &platform, &encoding, &language, &NameId);

	if ((NameId == TT_NAME_ID_FONT_FAMILY) || (NameId == TT_NAME_ID_PS_NAME)) {
	    if (isMacEnglish (platform, encoding, language)) {
		result = do_get_name (face, i, 0);
	    } else if (isUnicodeEnglish (platform, encoding, language)) {
		result = do_get_name (face, i, 1);
	    }

	    if (result != NULL) {
		if (NameId == TT_NAME_ID_FONT_FAMILY)
		    return result;
		else
		    ps_name = result;
	    }
	}
    }

    return ps_name ? ps_name : "unknown";
}

static void  add_mapping (struct xlfdmapping **listhead,
			  char *file_name,
			  const char *foundry_name,
			  const char *font_name,
			  const char *weight_name,
			  const char *widht_name,
			  TT_Face_Properties *properties,
			  char *encodingscheme
			  )
{
    struct xlfdmapping *p, *q;
    char char_for_spacing;
    
    /* allocate space for the file to xlfd mapping */
    p = malloc (sizeof (struct xlfdmapping));

    if (p == NULL) {
	exit (TTMKFS_MEM);
    }

    /*
     * we can't just write the mapping, but have to collect them.
     * To quote the mkfontdir man page: "The  first  line of fonts.dir
     * gives the number of fonts in the file. ... "fonts.scale" has
     * the same format as the "fonts.dir" file.". So we first have to
     * count and collect all our mappings.
     */

    if ((strncmp("jisx",encodingscheme,4)==0) ||
	(strncmp("gb2312",encodingscheme,5)==0) ||
	(strncmp("gbk",encodingscheme,3)==0) ||
	(strncmp("big5",encodingscheme,4)==0) ||
	(strncmp("ksc",encodingscheme,3)==0))
    {
	char_for_spacing = 'c';
    } else {
	char_for_spacing = (properties->postscript->isFixedPitch) ?
				'm' : 'p' ;
    }

    sprintf (p->mapping, "%s -%s-%s-%s-%c-%s--0-0-0-0-%c-0-%s\n",
	     file_name,
	     foundry_name,
	     font_name,
	     weight_name,
	     properties->os2->fsSelection & TT_SELECTION_ITALIC ? 'i' : 'r',
	     widht_name,
	     char_for_spacing,
	     encodingscheme);

    /* check that it is a new XLFD */
    for (q = *listhead; q ; q = q->next) {
	if (strcasecmp(p->mapping, q->mapping) == 0) {
	    free(p);
	    p=NULL;
	    break;
	}
    }

    if (p) {
	/* insert into our list */
	p->next = *listhead;
	*listhead = p;
	entries++;
    }
}

int
has_8bit_encoding (TT_CharMap *cmap, const int *table)
{
    int i;
    int missing = 0;

    for (i = 0; i < 224; i++) {
	if (table[i] < 1) 
	    continue;

	if (TT_Char_Index (*cmap, table[i]) == 0) {
	    missing++;
	}
    }

    return (missing > max_missing) ? 0 : 1;
}

static const char *
panose_get_weight (TT_Byte *panose)
{
    static char *panose_weight_latin[] = {
	"any",
	"no fit",
	"very light",
	"light",
	"thin",
	"book",
	"medium",
	"demi",
	"bold",
	"heavy",
	"black",
	"extra black",
    };

    switch (panose[TT_PANOSE_FAMILY_TYPE]) {
    case TT_PANOSE_FAMILY_LATIN_TEXT:
    case TT_PANOSE_FAMILY_LATIN_HAND:
    case TT_PANOSE_FAMILY_LATIN_DECORATIVE:
	if (!((panose[TT_PANOSE_WEIGHT] < TT_PANOSE_LT_WEIGHT_VERY_LIGHT) ||
	      (panose[TT_PANOSE_WEIGHT] > TT_PANOSE_LT_WEIGHT_EXTRA_BLACK))) {
	    return panose_weight_latin[panose[TT_PANOSE_WEIGHT]];
	}
	break;
    }

    return NULL;
}

static const char *
os2_get_weight (TT_UShort usWeightClass)
{
    switch (usWeightClass) {
    case TT_FW_THIN: return "thin";
    case TT_FW_EXTRALIGHT: return "extra light";
    case TT_FW_LIGHT: return "light";
    case TT_FW_NORMAL: return "medium";
    case TT_FW_MEDIUM: return "medium";
    case TT_FW_SEMIBOLD: return "semi bold";
    case TT_FW_BOLD: return "bold";
    case TT_FW_EXTRABOLD: return "extra bold";
    case TT_FW_BLACK:  return "black";
    }

    return NULL;
}

static const char *
get_weight (const char *name, TT_Face_Properties *properties)
{
    const char *result;
    
    if (use_panose && ((result = panose_get_weight (properties->os2->panose)) != NULL)) {
	return result;
    } else {
	if ((result = os2_get_weight (properties->os2->usWeightClass)) != NULL) {
	    return result;
	}
	
	fprintf (stderr, "font %s: unknown usWeightClass %hu\n", name, properties->os2->usWeightClass);
    }

    /* last fall back */
    return properties->os2->fsSelection & (1 << 5) ? "bold" : "medium";
}

static const char *
panose_get_width_name (TT_Byte *panose)
{
    static char *panose_spacing_latin_text [] = {
	"any",
	"no fit",
	"old style",
	"modern",
	"even width",
	"extended",
	"condensed",
	"very extended",
	"very condensed",
	"monospaced",
    };
    
    static char *panose_spacing_latin_decorative [] = {
	"any",
	"no fit",
	"super condensed",
	"very condensed",
	"condensed",
	"normal",
	"extended",
	"very extended",
	"super extended",
	"monospaced",
    };

    switch (panose[TT_PANOSE_FAMILY_TYPE]) {
    case TT_PANOSE_FAMILY_LATIN_TEXT:
	if (!((panose[TT_PANOSE_PROPORTION] < TT_PANOSE_LT_PROPORTION_OLD_STYLE) ||
	      (panose[TT_PANOSE_PROPORTION] > TT_PANOSE_LT_PROPORTION_MONOSPACED))) {
	    return panose_spacing_latin_text[panose[TT_PANOSE_PROPORTION]];
	}
	break;
	
    case TT_PANOSE_FAMILY_LATIN_DECORATIVE:
	if (!((panose[TT_PANOSE_ASPECT] < TT_PANOSE_LD_SUPER_CONDENSED) ||
	      (panose[TT_PANOSE_ASPECT] > TT_PANOSE_LD_MONOSPACED))) {
	    return panose_spacing_latin_decorative[panose[TT_PANOSE_ASPECT]];
	}
	break;
    }

    return NULL;
}

static const char *
os2_get_width_name (TT_UShort usWidthClass)
{
    switch (usWidthClass) {
    case TT_FWIDTH_ULTRA_CONDENSED: return "ultra condensed";
    case TT_FWIDTH_EXTRA_CONDENSED: return "extra condensed";
    case TT_FWIDTH_CONDENSED: return "condensed";
    case TT_FWIDTH_SEMI_CONDENSED: return "semi condensed";
    case TT_FWIDTH_NORMAL: return "normal";
    case TT_FWIDTH_SEMI_EXPANDED: return "semi expanded";
    case TT_FWIDTH_EXPANDED: return "expanded";
    case TT_FWIDTH_EXTRA_EXPANDED: return "extra expanded";
    case TT_FWIDTH_ULTRA_EXPANDED: return "ultra expanded";
    }

    return NULL;
}

static const char *
get_width_name (const char *name, TT_Face_Properties *properties)
{
    const char *result;
    
    if (use_panose && ((result = panose_get_width_name (properties->os2->panose)) != NULL)) {
	return result;
    } else {
	if ((result = os2_get_width_name (properties->os2->usWidthClass)) != NULL) {
	    return result;
	}
	
	fprintf (stderr, "font %s: unknown usWidthClass %hu\n", name, properties->os2->usWidthClass);
    }

    /* last fall back */
    return "normal";
}
